@@ -2,6 +2,7 @@ use std::io::{self, ErrorKind};
 use std::path::PathBuf;
 
 use crate::common::SudoPath;
+use crate::system::interface::ProcessId;
 use crate::{
     common::{context::LaunchType, Context},
     system::{Group, User},
@@ -14,9 +15,9 @@ pub trait RunOptions {
     fn chdir(&self) -> Option<&SudoPath>;
     fn is_login(&self) -> bool;
     fn user(&self) -> &User;
-    fn requesting_user(&self) -> &User;
     fn group(&self) -> &Group;
-    fn pid(&self) -> i32;
+    fn pid(&self) -> ProcessId;
+
     fn use_pty(&self) -> bool;
 }
 
@@ -49,15 +50,11 @@ impl RunOptions for Context {
         &self.target_user
     }
 
-    fn requesting_user(&self) -> &User {
-        &self.current_user
-    }
-
     fn group(&self) -> &Group {
         &self.target_group
     }
 
-    fn pid(&self) -> i32 {
+    fn pid(&self) -> ProcessId {
         self.process.pid
     }
 
