@@ -7,8 +7,7 @@ use crate::exec::event::{EventHandle, EventRegistry, PollEvent, Process, StopRea
 use crate::exec::use_pty::monitor::exec_monitor;
 use crate::exec::use_pty::SIGCONT_FG;
 use crate::exec::{
-    cond_fmt, handle_sigchld, opt_fmt, signal_fmt, terminate_process, ExecOutput, HandleSigchld,
-    ProcessOutput,
+    cond_fmt, handle_sigchld, signal_fmt, terminate_process, ExecOutput, HandleSigchld,
 };
 use crate::exec::{
     io_util::retry_while_interrupted,
@@ -22,7 +21,9 @@ use crate::system::signal::{
 };
 use crate::system::term::{Pty, PtyFollower, PtyLeader, TermSize, Terminal, UserTerm};
 use crate::system::wait::WaitOptions;
-use crate::system::{chown, fork, getpgrp, kill, killpg, FileCloser, ForkResult, Group, User};
+use crate::system::{
+    chown, fork, getpgrp, kill, killpg, FileCloser, ForkResult, Group, User, _exit,
+};
 use crate::system::{getpgid, interface::ProcessId};
 
 use super::pipe::Pipe;
@@ -32,7 +33,7 @@ pub(in crate::exec) fn exec_pty(
     sudo_pid: ProcessId,
     mut command: Command,
     user_tty: UserTerm,
-) -> io::Result<ProcessOutput> {
+) -> io::Result<ExecOutput> {
     // Allocate a pseudoterminal.
     let pty = get_pty()?;
 
@@ -162,7 +163,8 @@ pub(in crate::exec) fn exec_pty(
         }
     };
 
-    let ForkResult::Parent(monitor_pid) = fork().map_err(|err| {
+    // SAFETY: There should be no other threads at this point.
+    let ForkResult::Parent(monitor_pid) = (unsafe { fork() }).map_err(|err| {
         dev_error!("cannot fork monitor process: {err}");
         err
     })?
@@ -180,9 +182,9 @@ pub(in crate::exec) fn exec_pty(
             file_closer,
             original_set,
         ) {
-            Ok(exec_output) => return Ok(exec_output),
+            Ok(exec_output) => match exec_output {},
             Err(err) => {
-                // Disable nonblocking assetions as we will not poll the backchannel anymore.
+                // Disable nonblocking assertions as we will not poll the backchannel anymore.
                 backchannels.monitor.set_nonblocking_assertions(true);
 
                 match err.try_into() {
@@ -198,7 +200,8 @@ pub(in crate::exec) fn exec_pty(
             }
         }
 
-        return Ok(ProcessOutput::ChildExit);
+        // We call `_exit` instead of `exit` to avoid flushing the parent's IO streams by accident.
+        _exit(1);
     };
 
     // Close the file descriptors that we don't access
@@ -232,7 +235,7 @@ pub(in crate::exec) fn exec_pty(
         }
     }
 
-    let exit_reason = closure.run(&mut registry);
+    let exit_reason = closure.run(registry);
     // FIXME (ogsudo): Retry if `/dev/tty` is revoked.
 
     // Flush the terminal
@@ -251,11 +254,9 @@ pub(in crate::exec) fn exec_pty(
         }
     }
 
-    Ok(ProcessOutput::SudoExit {
-        output: ExecOutput {
-            command_exit_reason: exit_reason?,
-            restore_signal_handlers: Box::new(move || drop(closure.signal_handlers)),
-        },
+    Ok(ExecOutput {
+        command_exit_reason: exit_reason?,
+        restore_signal_handlers: Box::new(move || drop(closure.signal_handlers)),
     })
 }
 
@@ -349,7 +350,7 @@ impl ParentClosure {
         })
     }
 
-    fn run(&mut self, registry: &mut EventRegistry<Self>) -> io::Result<ExitReason> {
+    fn run(&mut self, registry: EventRegistry<Self>) -> io::Result<ExitReason> {
         match registry.event_loop(self) {
             StopReason::Break(err) | StopReason::Exit(ParentExit::Backchannel(err)) => Err(err),
             StopReason::Exit(ParentExit::Command(exit_reason)) => Ok(exit_reason),
@@ -432,7 +433,7 @@ impl ParentClosure {
     /// - is the same PID of the command, or
     /// - is in the process group of the command and either sudo or the command is the leader.
     fn is_self_terminating(&self, signaler_pid: ProcessId) -> bool {
-        if signaler_pid != 0 {
+        if signaler_pid.is_valid() {
             if Some(signaler_pid) == self.command_pid {
                 return true;
             }
@@ -616,12 +617,7 @@ impl ParentClosure {
             }
         };
 
-        dev_info!(
-            "parent received{} {} from {}",
-            opt_fmt(info.is_user_signaled(), " user signaled"),
-            info.signal(),
-            info.pid()
-        );
+        dev_info!("parent received{}", info);
 
         let Some(monitor_pid) = self.monitor_pid else {
             dev_info!("monitor was terminated, ignoring signal");
@@ -638,10 +634,17 @@ impl ParentClosure {
                     dev_warn!("cannot resize terminal: {}", err);
                 }
             }
-            // Skip the signal if it was sent by the user and it is self-terminating.
-            _ if info.is_user_signaled() && self.is_self_terminating(info.pid()) => {}
-            // FIXME: check `send_command_status`
-            signal => self.schedule_signal(signal, registry),
+            signal => {
+                if let Some(pid) = info.signaler_pid() {
+                    if self.is_self_terminating(pid) {
+                        // Skip the signal if it was sent by the user and it is self-terminating.
+                        return;
+                    }
+                }
+
+                // FIXME: check `send_command_status`
+                self.schedule_signal(signal, registry)
+            }
         }
     }
 
