@@ -2,7 +2,6 @@
 
 use std::{borrow::Cow, mem};
 
-use crate::common::context::{ContextAction, OptionsForContext};
 use crate::common::{SudoPath, SudoString};
 
 pub mod help;
@@ -733,7 +732,7 @@ fn ensure_is_absent(context: &str, thing: &dyn IsAbsent, name: &str) -> Result<(
     if thing.is_absent() {
         Ok(())
     } else {
-        Err(format!("{context} conflicts with {name}"))
+        Err(format!("{context} cannot be used together with {name}"))
     }
 }
 
@@ -795,98 +794,7 @@ fn reject_all(context: &str, opts: SudoOptions) -> Result<(), String> {
     }
 
     ensure_is_absent(context, &env_var_list, "environment variable")?;
-    ensure_is_absent(context, &positional_args, "positional argument")?;
+    ensure_is_absent(context, &positional_args, "command")?;
 
     Ok(())
 }
-
-impl From<SudoListOptions> for OptionsForContext {
-    fn from(opts: SudoListOptions) -> Self {
-        let SudoListOptions {
-            group,
-            non_interactive,
-            positional_args,
-            reset_timestamp,
-            stdin,
-            user,
-
-            list: _,
-            other_user: _,
-        } = opts;
-
-        Self {
-            action: ContextAction::List,
-
-            group,
-            non_interactive,
-            positional_args,
-            reset_timestamp,
-            stdin,
-            user,
-
-            chdir: None,
-            login: false,
-            shell: false,
-        }
-    }
-}
-
-impl From<SudoValidateOptions> for OptionsForContext {
-    fn from(opts: SudoValidateOptions) -> Self {
-        let SudoValidateOptions {
-            group,
-            non_interactive,
-            reset_timestamp,
-            stdin,
-            user,
-        } = opts;
-
-        Self {
-            action: ContextAction::Validate,
-
-            group,
-            non_interactive,
-            reset_timestamp,
-            stdin,
-            user,
-
-            chdir: None,
-            login: false,
-            positional_args: vec![],
-            shell: false,
-        }
-    }
-}
-
-impl From<SudoRunOptions> for OptionsForContext {
-    fn from(opts: SudoRunOptions) -> Self {
-        let SudoRunOptions {
-            chdir,
-            group,
-            login,
-            non_interactive,
-            positional_args,
-            reset_timestamp,
-            shell,
-            stdin,
-            user,
-
-            env_var_list: _,
-            preserve_env: _,
-        } = opts;
-
-        Self {
-            action: ContextAction::Run,
-
-            chdir,
-            group,
-            login,
-            non_interactive,
-            positional_args,
-            reset_timestamp,
-            shell,
-            stdin,
-            user,
-        }
-    }
-}
