@@ -1,9 +1,10 @@
 use crate::common::resolve::CurrentUser;
-use crate::common::{CommandAndArguments, Context, Environment};
+use crate::common::{CommandAndArguments, Context};
 use crate::sudo::{
     cli::{SudoAction, SudoRunOptions},
-    env::environment::get_target_environment,
+    env::environment::{get_target_environment, Environment},
 };
+use crate::system::interface::{GroupId, UserId};
 use crate::system::{Group, Hostname, Process, User};
 use std::collections::{HashMap, HashSet};
 
@@ -65,7 +66,7 @@ fn parse_env_commands(input: &str) -> Vec<(&str, Environment)> {
         .map(|e| {
             let (cmd, vars) = e.split_once('\n').unwrap();
 
-            let vars: Environment = vars
+            let vars = vars
                 .lines()
                 .map(|line| line.trim().split_once('=').unwrap())
                 .map(|(k, v)| (k.into(), v.into()))
@@ -82,8 +83,8 @@ fn create_test_context(sudo_options: &SudoRunOptions) -> Context {
         CommandAndArguments::build_from_args(None, sudo_options.positional_args.clone(), &path);
 
     let current_user = CurrentUser::fake(User {
-        uid: 1000,
-        gid: 1000,
+        uid: UserId::new(1000),
+        gid: GroupId::new(1000),
 
         name: "test".into(),
         gecos: String::new(),
@@ -94,13 +95,13 @@ fn create_test_context(sudo_options: &SudoRunOptions) -> Context {
     });
 
     let current_group = Group {
-        gid: 1000,
-        name: "test".to_string(),
+        gid: GroupId::new(1000),
+        name: Some("test".to_string()),
     };
 
     let root_user = User {
-        uid: 0,
-        gid: 0,
+        uid: UserId::ROOT,
+        gid: GroupId::new(0),
         name: "root".into(),
         gecos: String::new(),
         home: "/root".into(),
@@ -110,8 +111,8 @@ fn create_test_context(sudo_options: &SudoRunOptions) -> Context {
     };
 
     let root_group = Group {
-        gid: 0,
-        name: "root".to_string(),
+        gid: GroupId::new(0),
+        name: Some("root".to_string()),
     };
 
     Context {
@@ -135,13 +136,14 @@ fn create_test_context(sudo_options: &SudoRunOptions) -> Context {
         process: Process::new(),
         use_session_records: false,
         use_pty: true,
+        password_feedback: false,
     }
 }
 
 fn environment_to_set(environment: Environment) -> HashSet<String> {
     HashSet::from_iter(
         environment
-            .iter()
+            .into_iter()
             .map(|(k, v)| format!("{}={}", k.to_str().unwrap(), v.to_str().unwrap())),
     )
 }
@@ -157,10 +159,23 @@ fn test_environment_variable_filtering() {
             .try_into_run()
             .ok()
             .unwrap();
-        let settings = crate::sudoers::Judgement::default();
+        let settings = crate::defaults::Settings::default();
         let context = create_test_context(&options);
-        let resulting_env =
-            get_target_environment(initial_env.clone(), HashMap::new(), &context, &settings);
+        let resulting_env = get_target_environment(
+            initial_env.clone(),
+            HashMap::new(),
+            Vec::new(),
+            &context,
+            &crate::sudoers::Restrictions {
+                env_keep: settings.env_keep(),
+                env_check: settings.env_check(),
+                path: settings.secure_path(),
+                use_pty: true,
+                chdir: crate::sudoers::DirChange::Strict(None),
+                trust_environment: false,
+            },
+        )
+        .unwrap();
 
         let resulting_env = environment_to_set(resulting_env);
         let expected_env = environment_to_set(expected_env);
