@@ -1,84 +1,108 @@
-use std::ffi::OsStr;
+use std::ffi::{OsStr, OsString};
 use std::process::exit;
 
 use super::cli::{SudoRunOptions, SudoValidateOptions};
-use crate::common::context::OptionsForContext;
-use crate::common::resolve::CurrentUser;
-use crate::common::{Context, Environment, Error};
+use super::diagnostic;
+use crate::common::resolve::{AuthUser, CurrentUser};
+use crate::common::{Context, Error};
 use crate::exec::{ExecOutput, ExitReason};
 use crate::log::{auth_info, auth_warn};
 use crate::sudo::env::environment;
 use crate::sudo::Duration;
-use crate::sudoers::{Authorization, AuthorizationAllowed, DirChange, Policy, PreJudgementPolicy};
-use crate::system::interface::UserId;
+use crate::sudoers::{
+    AuthenticatingUser, Authentication, Authorization, DirChange, Judgement, Restrictions, Sudoers,
+};
 use crate::system::term::current_tty_name;
 use crate::system::timestamp::{RecordScope, SessionRecordFile, TouchResult};
 use crate::system::{escape_os_str_lossy, Process};
 
 mod list;
 
-pub trait PolicyPlugin {
-    type PreJudgementPolicy: PreJudgementPolicy;
-    type Policy: Policy;
-
-    fn init(&mut self) -> Result<Self::PreJudgementPolicy, Error>;
-    fn judge(
-        &mut self,
-        pre: Self::PreJudgementPolicy,
-        context: &Context,
-    ) -> Result<Self::Policy, Error>;
-}
-
 pub trait AuthPlugin {
-    fn init(&mut self, context: &Context) -> Result<(), Error>;
+    fn init(&mut self, context: &Context, auth_user: AuthUser) -> Result<(), Error>;
     fn authenticate(&mut self, non_interactive: bool, max_tries: u16) -> Result<(), Error>;
-    fn pre_exec(&mut self, target_user: &str) -> Result<Environment, Error>;
+    fn pre_exec(&mut self, target_user: &str) -> Result<Vec<(OsString, OsString)>, Error>;
     fn cleanup(&mut self);
 }
 
-pub struct Pipeline<Policy: PolicyPlugin, Auth: AuthPlugin> {
-    pub policy: Policy,
+pub struct Pipeline<Auth: AuthPlugin> {
     pub authenticator: Auth,
 }
 
-impl<Policy: PolicyPlugin, Auth: AuthPlugin> Pipeline<Policy, Auth> {
-    pub fn run(mut self, cmd_opts: SudoRunOptions) -> Result<(), Error> {
-        if !cmd_opts.env_var_list.is_empty() {
-            eprintln_ignore_io_error!(
-                "warning: CLI-level env var list has not yet been implemented and will be ignored"
-            )
-        }
+fn read_sudoers() -> Result<Sudoers, Error> {
+    let sudoers_path = super::candidate_sudoers_file();
+
+    let (sudoers, syntax_errors) =
+        Sudoers::open(sudoers_path).map_err(|e| Error::Configuration(format!("{e}")))?;
+
+    for crate::sudoers::Error {
+        source,
+        location,
+        message,
+    } in syntax_errors
+    {
+        let path = source.as_deref().unwrap_or(sudoers_path);
+        diagnostic::diagnostic!("{message}", path @ location);
+    }
+
+    Ok(sudoers)
+}
+
+fn judge(mut policy: Sudoers, context: &Context) -> Result<Judgement, Error> {
+    Ok(policy.check(
+        &*context.current_user,
+        &context.hostname,
+        crate::sudoers::Request {
+            user: &context.target_user,
+            group: &context.target_group,
+            command: &context.command.command,
+            arguments: &context.command.arguments,
+        },
+    ))
+}
+
+impl<Auth: AuthPlugin> Pipeline<Auth> {
+    pub fn run(mut self, mut cmd_opts: SudoRunOptions) -> Result<(), Error> {
+        let mut policy = read_sudoers()?;
+
+        let user_requested_env_vars = std::mem::take(&mut cmd_opts.env_var_list);
+
         if !cmd_opts.preserve_env.is_nothing() {
             eprintln_ignore_io_error!(
                 "warning: `--preserve-env` has not yet been implemented and will be ignored"
             )
         }
 
-        let pre = self.policy.init()?;
-        let mut context = build_context(cmd_opts.into(), &pre)?;
+        let mut context = Context::from_run_opts(cmd_opts, &mut policy)?;
 
-        let policy = self.policy.judge(pre, &context)?;
-        let authorization = policy.authorization();
+        let policy = judge(policy, &context)?;
 
-        match authorization {
-            Authorization::Forbidden => {
-                return Err(Error::auth(&format!(
-                    "I'm sorry {}. I'm afraid I can't do that",
-                    context.current_user.name
-                )));
-            }
-            Authorization::Allowed(auth) => {
-                self.apply_policy_to_context(&mut context, &policy)?;
-                self.auth_and_update_record_file(&context, auth)?;
-            }
-        }
+        let Authorization::Allowed(auth, controls) = policy.authorization() else {
+            return Err(Error::Authorization(context.current_user.name.to_string()));
+        };
 
-        let additional_env = self.authenticator.pre_exec(&context.target_user.name)?;
+        self.apply_policy_to_context(&mut context, &auth, &controls)?;
+        self.auth_and_update_record_file(&mut context, &auth)?;
 
         // build environment
-        let current_env = std::env::vars_os().collect();
-        let target_env =
-            environment::get_target_environment(current_env, additional_env, &context, &policy);
+        let additional_env = self.authenticator.pre_exec(&context.target_user.name)?;
+
+        let current_env = environment::system_environment();
+        let (checked_vars, trusted_vars) = if controls.trust_environment {
+            (vec![], user_requested_env_vars)
+        } else {
+            (user_requested_env_vars, vec![])
+        };
+
+        let mut target_env = environment::get_target_environment(
+            current_env,
+            additional_env,
+            checked_vars,
+            &context,
+            &controls,
+        )?;
+
+        environment::dangerous_extend(&mut target_env, trusted_vars);
 
         let pid = context.process.pid;
 
@@ -113,18 +137,16 @@ impl<Policy: PolicyPlugin, Auth: AuthPlugin> Pipeline<Policy, Auth> {
     }
 
     pub fn run_validate(mut self, cmd_opts: SudoValidateOptions) -> Result<(), Error> {
-        let pre = self.policy.init()?;
-        let context = build_context(cmd_opts.into(), &pre)?;
+        let policy = read_sudoers()?;
 
-        match pre.validate_authorization() {
+        let mut context = Context::from_validate_opts(cmd_opts)?;
+
+        match policy.validate_authorization() {
             Authorization::Forbidden => {
-                return Err(Error::auth(&format!(
-                    "I'm sorry {}. I'm afraid I can't do that",
-                    context.current_user.name
-                )));
+                return Err(Error::Authorization(context.current_user.name.to_string()));
             }
-            Authorization::Allowed(auth) => {
-                self.auth_and_update_record_file(&context, auth)?;
+            Authorization::Allowed(auth, ()) => {
+                self.auth_and_update_record_file(&mut context, &auth)?;
             }
         }
 
@@ -133,23 +155,32 @@ impl<Policy: PolicyPlugin, Auth: AuthPlugin> Pipeline<Policy, Auth> {
 
     fn auth_and_update_record_file(
         &mut self,
-        context: &Context,
-        AuthorizationAllowed {
+        context: &mut Context,
+        &Authentication {
             must_authenticate,
             prior_validity,
             allowed_attempts,
-        }: AuthorizationAllowed,
+            ref credential,
+            ..
+        }: &Authentication,
     ) -> Result<(), Error> {
         let scope = RecordScope::for_process(&Process::new());
         let mut auth_status = determine_auth_status(
             must_authenticate,
             context.use_session_records,
             scope,
-            context.current_user.uid,
             &context.current_user,
             prior_validity,
         );
-        self.authenticator.init(context)?;
+
+        let auth_user = match credential {
+            AuthenticatingUser::InvokingUser => {
+                AuthUser::from_current_user(context.current_user.clone())
+            }
+            AuthenticatingUser::Root => AuthUser::resolve_root_for_rootpw()?,
+        };
+
+        self.authenticator.init(context, auth_user)?;
         if auth_status.must_authenticate {
             self.authenticator
                 .authenticate(context.non_interactive, allowed_attempts)?;
@@ -169,10 +200,11 @@ impl<Policy: PolicyPlugin, Auth: AuthPlugin> Pipeline<Policy, Auth> {
     fn apply_policy_to_context(
         &mut self,
         context: &mut Context,
-        policy: &<Policy as PolicyPlugin>::Policy,
+        auth: &Authentication,
+        controls: &Restrictions,
     ) -> Result<(), crate::common::Error> {
         // see if the chdir flag is permitted
-        match policy.chdir() {
+        match controls.chdir {
             DirChange::Any => {}
             DirChange::Strict(optdir) => {
                 if let Some(chdir) = &context.chdir {
@@ -191,32 +223,21 @@ impl<Policy: PolicyPlugin, Auth: AuthPlugin> Pipeline<Policy, Auth> {
             context.chdir = Some(dir.expand_tilde_in_path(&context.target_user.name)?)
         }
 
-        // override the default pty behaviour if indicated
-        if !policy.use_pty() {
-            context.use_pty = false
-        }
+        // in case the user could set these from the commandline, something more fancy
+        // could be needed, but here we copy these -- perhaps we should split up the Context type
+        context.use_pty = controls.use_pty;
+        context.password_feedback = auth.pwfeedback;
 
         Ok(())
     }
 }
 
-fn build_context(
-    cmd_opts: OptionsForContext,
-    pre: &dyn PreJudgementPolicy,
-) -> Result<Context, Error> {
-    let secure_path: String = pre
-        .secure_path()
-        .unwrap_or_else(|| std::env::var("PATH").unwrap_or_default());
-    Context::build_from_options(cmd_opts, secure_path)
-}
-
 /// This should determine what the authentication status for the given record
 /// match limit and origin/target user from the context is.
 fn determine_auth_status(
     must_policy_authenticate: bool,
     use_session_records: bool,
     record_for: Option<RecordScope>,
-    auth_uid: UserId,
     current_user: &CurrentUser,
     prior_validity: Duration,
 ) -> AuthStatus {
@@ -225,7 +246,7 @@ fn determine_auth_status(
     } else if let (true, Some(record_for)) = (use_session_records, record_for) {
         match SessionRecordFile::open_for_user(current_user, prior_validity) {
             Ok(mut sr) => {
-                match sr.touch(record_for, auth_uid) {
+                match sr.touch(record_for, current_user.uid) {
                     // if a record was found and updated within the timeout, we do not need to authenticate
                     Ok(TouchResult::Updated { .. }) => AuthStatus::new(false, Some(sr)),
                     Ok(TouchResult::NotFound | TouchResult::Outdated { .. }) => {
