@@ -6,9 +6,11 @@ use crate::common::{
     HARDENED_ENUM_VALUE_0, HARDENED_ENUM_VALUE_1, HARDENED_ENUM_VALUE_2, HARDENED_ENUM_VALUE_3,
     HARDENED_ENUM_VALUE_4,
 };
+use crate::defaults;
 
 /// The Sudoers file allows negating items with the exclamation mark.
-#[cfg_attr(test, derive(Debug, PartialEq, Eq))]
+#[cfg_attr(test, derive(Debug, Eq))]
+#[derive(Clone, PartialEq)]
 #[repr(u32)]
 pub enum Qualified<T> {
     Allow(T) = HARDENED_ENUM_VALUE_0,
@@ -16,20 +18,6 @@ pub enum Qualified<T> {
 }
 
 impl<T> Qualified<T> {
-    pub fn as_ref(&self) -> Qualified<&T> {
-        match self {
-            Qualified::Allow(item) => Qualified::Allow(item),
-            Qualified::Forbid(item) => Qualified::Forbid(item),
-        }
-    }
-
-    pub fn negate(&self) -> Qualified<&T> {
-        match self {
-            Qualified::Allow(item) => Qualified::Forbid(item),
-            Qualified::Forbid(item) => Qualified::Allow(item),
-        }
-    }
-
     #[cfg(test)]
     pub fn as_allow(&self) -> Option<&T> {
         if let Self::Allow(v) = self {
@@ -44,19 +32,37 @@ impl<T> Qualified<T> {
 pub type Spec<T> = Qualified<Meta<T>>;
 pub type SpecList<T> = Vec<Spec<T>>;
 
+/// A generic mapping function (only used for turning `Spec<SimpleCommand>` into `Spec<Command>`)
+impl<T> Spec<T> {
+    pub fn map<U>(self, f: impl Fn(T) -> U) -> Spec<U> {
+        let transform = |meta| match meta {
+            Meta::All => Meta::All,
+            Meta::Alias(alias) => Meta::Alias(alias),
+            Meta::Only(x) => Meta::Only(f(x)),
+        };
+
+        match self {
+            Qualified::Allow(x) => Qualified::Allow(transform(x)),
+            Qualified::Forbid(x) => Qualified::Forbid(transform(x)),
+        }
+    }
+}
+
 /// An identifier is a name or a #number
 #[cfg_attr(test, derive(Clone, Debug, PartialEq, Eq))]
+#[repr(u32)]
 pub enum Identifier {
-    Name(SudoString),
-    ID(u32),
+    Name(SudoString) = HARDENED_ENUM_VALUE_0,
+    ID(u32) = HARDENED_ENUM_VALUE_1,
 }
 
 /// A userspecifier is either a username, or a (non-unix) group name, or netgroup
 #[cfg_attr(test, derive(Clone, Debug, PartialEq, Eq))]
+#[repr(u32)]
 pub enum UserSpecifier {
-    User(Identifier),
-    Group(Identifier),
-    NonunixGroup(Identifier),
+    User(Identifier) = HARDENED_ENUM_VALUE_0,
+    Group(Identifier) = HARDENED_ENUM_VALUE_1,
+    NonunixGroup(Identifier) = HARDENED_ENUM_VALUE_2,
 }
 
 /// The RunAs specification consists of a (possibly empty) list of userspecifiers, followed by a (possibly empty) list of groups.
@@ -78,18 +84,35 @@ pub enum Authenticate {
     Nopasswd = HARDENED_ENUM_VALUE_2,
 }
 
+#[derive(Copy, Clone, Default, PartialEq)]
+#[cfg_attr(test, derive(Debug, Eq))]
+#[repr(u32)]
+pub enum EnvironmentControl {
+    #[default]
+    Implicit = HARDENED_ENUM_VALUE_0,
+    // PASSWD:
+    Setenv = HARDENED_ENUM_VALUE_1,
+    // NOPASSWD:
+    Nosetenv = HARDENED_ENUM_VALUE_2,
+}
+
 /// Commands in /etc/sudoers can have attributes attached to them, such as NOPASSWD, NOEXEC, ...
 #[derive(Default, Clone, PartialEq)]
 #[cfg_attr(test, derive(Debug, Eq))]
 pub struct Tag {
-    pub authenticate: Authenticate,
-    pub cwd: Option<ChDir>,
+    pub(super) authenticate: Authenticate,
+    pub(super) cwd: Option<ChDir>,
+    pub(super) env: EnvironmentControl,
 }
 
 impl Tag {
     pub fn needs_passwd(&self) -> bool {
         matches!(self.authenticate, Authenticate::None | Authenticate::Passwd)
     }
+
+    pub fn allows_setenv(&self) -> bool {
+        matches!(self.env, EnvironmentControl::Setenv)
+    }
 }
 
 /// Commands with attached attributes.
@@ -113,24 +136,19 @@ pub enum Directive {
     HostAlias(Defs<Hostname>) = HARDENED_ENUM_VALUE_1,
     CmndAlias(Defs<Command>) = HARDENED_ENUM_VALUE_2,
     RunasAlias(Defs<UserSpecifier>) = HARDENED_ENUM_VALUE_3,
-    Defaults(Vec<(String, ConfigValue)>) = HARDENED_ENUM_VALUE_4,
-}
-
-pub type TextEnum = crate::defaults::StrEnum<'static>;
-
-pub enum ConfigValue {
-    Flag(bool),
-    Text(Option<Box<str>>),
-    Num(i64),
-    List(Mode, Vec<String>),
-    Enum(TextEnum),
+    Defaults(Vec<defaults::SettingsModifier>, ConfigScope) = HARDENED_ENUM_VALUE_4,
 }
 
+/// AST object for the 'context' (host, user, cmnd, runas) of a Defaults directive
 #[repr(u32)]
-pub enum Mode {
-    Add = HARDENED_ENUM_VALUE_0,
-    Set = HARDENED_ENUM_VALUE_1,
-    Del = HARDENED_ENUM_VALUE_2,
+pub enum ConfigScope {
+    // "Defaults entries are parsed in the following order:
+    // generic, host and user Defaults first, then runas Defaults and finally command defaults."
+    Generic = HARDENED_ENUM_VALUE_0,
+    Host(SpecList<Hostname>) = HARDENED_ENUM_VALUE_1,
+    User(SpecList<UserSpecifier>) = HARDENED_ENUM_VALUE_2,
+    RunAs(SpecList<UserSpecifier>) = HARDENED_ENUM_VALUE_3,
+    Command(SpecList<SimpleCommand>) = HARDENED_ENUM_VALUE_4,
 }
 
 /// The Sudoers file can contain permissions and directives
@@ -138,8 +156,8 @@ pub enum Mode {
 pub enum Sudo {
     Spec(PermissionSpec) = HARDENED_ENUM_VALUE_0,
     Decl(Directive) = HARDENED_ENUM_VALUE_1,
-    Include(String) = HARDENED_ENUM_VALUE_2,
-    IncludeDir(String) = HARDENED_ENUM_VALUE_3,
+    Include(String, Span) = HARDENED_ENUM_VALUE_2,
+    IncludeDir(String, Span) = HARDENED_ENUM_VALUE_3,
     LineComment = HARDENED_ENUM_VALUE_4,
 }
 
@@ -171,7 +189,7 @@ impl Sudo {
 
     #[cfg(test)]
     pub fn as_include(&self) -> &str {
-        if let Self::Include(v) = self {
+        if let Self::Include(v, _) = self {
             v
         } else {
             panic!()
@@ -193,10 +211,9 @@ impl Sudo {
 /// identifier = name
 ///            | #<numerical id>
 /// ```
-
 impl Parse for Identifier {
-    fn parse(stream: &mut impl CharStream) -> Parsed<Self> {
-        if accept_if(|c| c == '#', stream).is_some() {
+    fn parse(stream: &mut CharStream) -> Parsed<Self> {
+        if stream.eat_char('#') {
             let Digits(guid) = expect_nonterminal(stream)?;
             make(Identifier::ID(guid))
         } else {
@@ -215,9 +232,8 @@ impl Many for Identifier {}
 ///
 /// This computes the correct negation with multiple exclamation marks in the parsing stage so we
 /// are not bothered by it later.
-
 impl<T: Parse + UserFriendly> Parse for Qualified<T> {
-    fn parse(stream: &mut impl CharStream) -> Parsed<Self> {
+    fn parse(stream: &mut CharStream) -> Parsed<Self> {
         if is_syntax('!', stream)? {
             let mut neg = true;
             while is_syntax('!', stream)? {
@@ -242,9 +258,8 @@ impl<T: Many> Many for Qualified<T> {
 }
 
 /// Helper function for parsing `Meta<T>` things where T is not a token
-
 fn parse_meta<T: Parse>(
-    stream: &mut impl CharStream,
+    stream: &mut CharStream,
     embed: impl FnOnce(SudoString) -> T,
 ) -> Parsed<Meta<T>> {
     if let Some(meta) = try_nonterminal(stream)? {
@@ -259,9 +274,8 @@ fn parse_meta<T: Parse>(
 }
 
 /// Since Identifier is not a token, add the parser for `Meta<Identifier>`
-
 impl Parse for Meta<Identifier> {
-    fn parse(stream: &mut impl CharStream) -> Parsed<Self> {
+    fn parse(stream: &mut CharStream) -> Parsed<Self> {
         parse_meta(stream, Identifier::Name)
     }
 }
@@ -274,16 +288,16 @@ impl Parse for Meta<Identifier> {
 ///          | +netgroup
 /// ```
 impl Parse for UserSpecifier {
-    fn parse(stream: &mut impl CharStream) -> Parsed<Self> {
-        let userspec = if accept_if(|c| c == '%', stream).is_some() {
-            let ctor = if accept_if(|c| c == ':', stream).is_some() {
+    fn parse(stream: &mut CharStream) -> Parsed<Self> {
+        let userspec = if stream.eat_char('%') {
+            let ctor = if stream.eat_char(':') {
                 UserSpecifier::NonunixGroup
             } else {
                 UserSpecifier::Group
             };
             // in this case we must fail 'hard', since input has been consumed
             ctor(expect_nonterminal(stream)?)
-        } else if accept_if(|c| c == '+', stream).is_some() {
+        } else if stream.eat_char('+') {
             // TODO Netgroups
             unrecoverable!(stream, "netgroups are not supported yet");
         } else {
@@ -299,7 +313,7 @@ impl Many for UserSpecifier {}
 
 /// UserSpecifier is not a token, implement the parser for `Meta<UserSpecifier>`
 impl Parse for Meta<UserSpecifier> {
-    fn parse(stream: &mut impl CharStream) -> Parsed<Self> {
+    fn parse(stream: &mut CharStream) -> Parsed<Self> {
         parse_meta(stream, |name| UserSpecifier::User(Identifier::Name(name)))
     }
 }
@@ -309,7 +323,7 @@ impl Parse for Meta<UserSpecifier> {
 /// runas = "(", userlist, (":", grouplist?)?, ")"
 /// ```
 impl Parse for RunAs {
-    fn parse(stream: &mut impl CharStream) -> Parsed<Self> {
+    fn parse(stream: &mut CharStream) -> Parsed<Self> {
         try_syntax('(', stream)?;
         let users = try_nonterminal(stream).unwrap_or_default();
         let groups = maybe(try_syntax(':', stream).and_then(|_| try_nonterminal(stream)))?
@@ -336,8 +350,10 @@ pub type Modifier = Box<dyn Fn(&mut Tag)>;
 // to be more general, we impl Parse for Meta<Tag> so a future tag like "AFOOBAR" can be added with no problem
 
 impl Parse for MetaOrTag {
-    fn parse(stream: &mut impl CharStream) -> Parsed<Self> {
+    fn parse(stream: &mut CharStream) -> Parsed<Self> {
         use Meta::*;
+
+        let start_pos = stream.get_pos();
         let AliasName(keyword) = try_nonterminal(stream)?;
 
         let mut switch = |modifier: fn(&mut Tag)| {
@@ -346,13 +362,47 @@ impl Parse for MetaOrTag {
         };
 
         let result: Modifier = match keyword.as_str() {
+            // we do not support these, and that should make sudo-rs "fail safe"
+            "INTERCEPT" | "NOEXEC" => unrecoverable!(
+                pos = start_pos,
+                stream,
+                "NOEXEC and INTERCEPT are not supported by sudo-rs"
+            ),
+            // this is less fatal
+            "LOG_INPUT" | "NOLOG_INPUT" | "LOG_OUTPUT" | "NOLOG_OUTPUT" | "MAIL" | "NOMAIL" => {
+                eprintln_ignore_io_error!(
+                    "warning: {} tags are ignored by sudo-rs",
+                    keyword.as_str()
+                );
+                switch(|_| {})?
+            }
+
+            // 'FOLLOW' and 'NOFOLLOW' are only usable in a sudoedit context, which will result in
+            // a parse error elsewhere. 'EXEC' and 'NOINTERCEPT' are the default behaviour.
+            "FOLLOW" | "NOFOLLOW" | "EXEC" | "NOINTERCEPT" => switch(|_| {})?,
+
+            "SETENV" => switch(|tag| tag.env = EnvironmentControl::Setenv)?,
+            "NOSETENV" => switch(|tag| tag.env = EnvironmentControl::Nosetenv)?,
             "PASSWD" => switch(|tag| tag.authenticate = Authenticate::Passwd)?,
             "NOPASSWD" => switch(|tag| tag.authenticate = Authenticate::Nopasswd)?,
+
             "CWD" => {
                 expect_syntax('=', stream)?;
                 let path: ChDir = expect_nonterminal(stream)?;
                 Box::new(move |tag| tag.cwd = Some(path.clone()))
             }
+            // we do not support these, and that should make sudo-rs "fail safe"
+            spec @ ("CHROOT" | "TIMEOUT" | "NOTBEFORE" | "NOTAFTER") => unrecoverable!(
+                pos = start_pos,
+                stream,
+                "{spec} is not supported by sudo-rs"
+            ),
+            "ROLE" | "TYPE" => unrecoverable!(
+                pos = start_pos,
+                stream,
+                "role based access control is not yet supported by sudo-rs"
+            ),
+
             "ALL" => return make(MetaOrTag(All)),
             alias => return make(MetaOrTag(Alias(alias.to_string()))),
         };
@@ -365,9 +415,8 @@ impl Parse for MetaOrTag {
 /// ```text
 /// commandspec = [tag modifiers]*, command
 /// ```
-
 impl Parse for CommandSpec {
-    fn parse(stream: &mut impl CharStream) -> Parsed<Self> {
+    fn parse(stream: &mut CharStream) -> Parsed<Self> {
         let mut tags = vec![];
         while let Some(MetaOrTag(keyword)) = try_nonterminal(stream)? {
             use Qualified::Allow;
@@ -414,9 +463,8 @@ impl Parse for CommandSpec {
 /// ```text
 /// (host,runas,commandspec) = hostlist, "=", [runas?, commandspec]+
 /// ```
-
 impl Parse for (SpecList<Hostname>, Vec<(Option<RunAs>, CommandSpec)>) {
-    fn parse(stream: &mut impl CharStream) -> Parsed<Self> {
+    fn parse(stream: &mut CharStream) -> Parsed<Self> {
         let hosts = try_nonterminal(stream)?;
         expect_syntax('=', stream)?;
         let runas_cmds = expect_nonterminal(stream)?;
@@ -437,7 +485,7 @@ impl Many for (SpecList<Hostname>, Vec<(Option<RunAs>, CommandSpec)>) {
 /// (runas,commandspec) = runas?, commandspec
 /// ```
 impl Parse for (Option<RunAs>, CommandSpec) {
-    fn parse(stream: &mut impl CharStream) -> Parsed<Self> {
+    fn parse(stream: &mut CharStream) -> Parsed<Self> {
         let runas: Option<RunAs> = try_nonterminal(stream)?;
         let cmd = if runas.is_some() {
             expect_nonterminal(stream)?
@@ -463,14 +511,13 @@ impl Many for (Option<RunAs>, CommandSpec) {}
 /// ```
 /// There is a syntactical ambiguity in the sudoer Directive and Permission specifications, so we
 /// have to parse them 'together' and do a delayed decision on which category we are in.
-
 impl Parse for Sudo {
     // note: original sudo would reject:
     //   "User_Alias, user machine = command"
     // but accept:
     //   "user, User_Alias machine = command"; this does the same
-    fn parse(stream: &mut impl CharStream) -> Parsed<Sudo> {
-        if accept_if(|c| c == '@', stream).is_some() {
+    fn parse(stream: &mut CharStream) -> Parsed<Sudo> {
+        if stream.eat_char('@') {
             return parse_include(stream);
         }
 
@@ -493,7 +540,7 @@ impl Parse for Sudo {
                 // the failed "try_nonterminal::<Identifier>" will have consumed the '#'
                 // the most ignominious part of sudoers: having to parse bits of comments
                 parse_include(stream).or_else(|_| {
-                    while accept_if(|c| c != '\n', stream).is_some() {}
+                    stream.skip_to_newline();
                     make(Sudo::LineComment)
                 })
             };
@@ -503,7 +550,7 @@ impl Parse for Sudo {
         if let Some(users) = maybe(try_nonterminal::<SpecList<_>>(stream))? {
             // element 1 always exists (parse_list fails on an empty list)
             let key = &users[0];
-            if let Some(directive) = maybe(get_directive(key, stream))? {
+            if let Some(directive) = maybe(get_directive(key, stream, start_pos))? {
                 if users.len() != 1 {
                     unrecoverable!(pos = start_pos, stream, "invalid user name list");
                 }
@@ -520,13 +567,12 @@ impl Parse for Sudo {
 }
 
 /// Parse the include/include dir part that comes after the '#' or '@' prefix symbol
-
-fn parse_include(stream: &mut impl CharStream) -> Parsed<Sudo> {
-    fn get_path(stream: &mut impl CharStream) -> Parsed<String> {
-        if accept_if(|c| c == '"', stream).is_some() {
+fn parse_include(stream: &mut CharStream) -> Parsed<Sudo> {
+    fn get_path(stream: &mut CharStream, key_pos: (usize, usize)) -> Parsed<(String, Span)> {
+        let path = if stream.eat_char('"') {
             let QuotedInclude(path) = expect_nonterminal(stream)?;
             expect_syntax('"', stream)?;
-            make(path)
+            path
         } else {
             let value_pos = stream.get_pos();
             let IncludePath(path) = expect_nonterminal(stream)?;
@@ -537,23 +583,33 @@ fn parse_include(stream: &mut impl CharStream) -> Parsed<Sudo> {
                     "use quotes around filenames or escape whitespace"
                 )
             }
-            make(path)
-        }
+            path
+        };
+        make((
+            path,
+            Span {
+                start: key_pos,
+                end: stream.get_pos(),
+            },
+        ))
     }
 
     let key_pos = stream.get_pos();
     let result = match try_nonterminal(stream)? {
-        Some(Username(key)) if key == "include" => Sudo::Include(get_path(stream)?),
-        Some(Username(key)) if key == "includedir" => Sudo::IncludeDir(get_path(stream)?),
+        Some(Username(key)) if key == "include" => {
+            let (path, span) = get_path(stream, key_pos)?;
+            Sudo::Include(path, span)
+        }
+        Some(Username(key)) if key == "includedir" => {
+            let (path, span) = get_path(stream, key_pos)?;
+            Sudo::IncludeDir(path, span)
+        }
         _ => unrecoverable!(pos = key_pos, stream, "unknown directive"),
     };
 
     make(result)
 }
 
-use crate::defaults::sudo_default;
-use crate::defaults::SudoDefault as Setting;
-
 /// grammar:
 /// ```text
 /// name = definition [ : name = definiton [ : ... ] ]
@@ -564,7 +620,7 @@ where
     T: UserFriendly,
     Meta<T>: Parse + Many,
 {
-    fn parse(stream: &mut impl CharStream) -> Parsed<Self> {
+    fn parse(stream: &mut CharStream) -> Parsed<Self> {
         let begin_pos = stream.get_pos();
         let AliasName(name) = try_nonterminal(stream)?;
         if name == "ALL" {
@@ -584,9 +640,15 @@ impl<T> Many for Def<T> {
     const SEP: char = ':';
 }
 
+// NOTE: This function is a bit of a hack, since it relies on the fact that all directives
+// occur in the spot of a username, and are of a form that would otherwise be a legal user name.
+// I.e. after a valid username has been parsed, we check if it isn't actually a valid start of a
+// directive. A more robust solution would be to use the approach taken by the `MetaOrTag` above.
+
 fn get_directive(
     perhaps_keyword: &Spec<UserSpecifier>,
-    stream: &mut impl CharStream,
+    stream: &mut CharStream,
+    begin_pos: (usize, usize),
 ) -> Parsed<Directive> {
     use super::ast::Directive::*;
     use super::ast::Meta::*;
@@ -601,7 +663,30 @@ fn get_directive(
         "Host_Alias" => make(HostAlias(expect_nonterminal(stream)?)),
         "Cmnd_Alias" | "Cmd_Alias" => make(CmndAlias(expect_nonterminal(stream)?)),
         "Runas_Alias" => make(RunasAlias(expect_nonterminal(stream)?)),
-        "Defaults" => make(Defaults(expect_nonterminal(stream)?)),
+        "Defaults" => {
+            //HACK: this avoids having to add "Defaults@" etc as separate tokens; but relying
+            //on positional information during parsing is of course, cheating.
+            let allow_scope_modifier = stream.get_pos().0 == begin_pos.0
+                && stream.get_pos().1 - begin_pos.1 == "Defaults".len();
+
+            let scope = if allow_scope_modifier {
+                if is_syntax('@', stream)? {
+                    ConfigScope::Host(expect_nonterminal(stream)?)
+                } else if is_syntax(':', stream)? {
+                    ConfigScope::User(expect_nonterminal(stream)?)
+                } else if is_syntax('!', stream)? {
+                    ConfigScope::Command(expect_nonterminal(stream)?)
+                } else if is_syntax('>', stream)? {
+                    ConfigScope::RunAs(expect_nonterminal(stream)?)
+                } else {
+                    ConfigScope::Generic
+                }
+            } else {
+                ConfigScope::Generic
+            };
+
+            make(Defaults(expect_nonterminal(stream)?, scope))
+        }
         _ => reject(),
     }
 }
@@ -610,13 +695,13 @@ fn get_directive(
 /// ```text
 /// parameter = name [+-]?= ...
 /// ```
-impl Parse for (String, ConfigValue) {
-    fn parse(stream: &mut impl CharStream) -> Parsed<Self> {
+impl Parse for defaults::SettingsModifier {
+    fn parse(stream: &mut CharStream) -> Parsed<Self> {
         let id_pos = stream.get_pos();
 
         // Parse multiple entries enclosed in quotes (for list-like Defaults-settings)
-        let parse_vars = |stream: &mut _| -> Parsed<Vec<String>> {
-            if accept_if(|c| c == '"', stream).is_some() {
+        let parse_vars = |stream: &mut CharStream| -> Parsed<Vec<String>> {
+            if stream.eat_char('"') {
                 let mut result = Vec::new();
                 while let Some(EnvVar(name)) = try_nonterminal(stream)? {
                     result.push(name);
@@ -642,18 +727,19 @@ impl Parse for (String, ConfigValue) {
         };
 
         // Parse the remainder of a list variable
-        let list_items = |mode: Mode, name: String, cfg: Setting, stream: &mut _| {
-            expect_syntax('=', stream)?;
-            if !matches!(cfg, Setting::List(_)) {
-                unrecoverable!(pos = id_pos, stream, "{name} is not a list parameter");
-            }
+        let list_items =
+            |mode: defaults::ListMode, name: String, cfg: defaults::SettingKind, stream: &mut _| {
+                expect_syntax('=', stream)?;
+                let defaults::SettingKind::List(checker) = cfg else {
+                    unrecoverable!(pos = id_pos, stream, "{name} is not a list parameter");
+                };
 
-            make((name, ConfigValue::List(mode, parse_vars(stream)?)))
-        };
+                make(checker(mode, parse_vars(stream)?))
+            };
 
         // Parse a text parameter
-        let text_item = |stream: &mut _| {
-            if accept_if(|c| c == '"', stream).is_some() {
+        let text_item = |stream: &mut CharStream| {
+            if stream.eat_char('"') {
                 let QuotedText(text) = expect_nonterminal(stream)?;
                 expect_syntax('"', stream)?;
                 make(text)
@@ -663,54 +749,42 @@ impl Parse for (String, ConfigValue) {
             }
         };
 
-        use crate::defaults::OptTuple;
-
         if is_syntax('!', stream)? {
             let value_pos = stream.get_pos();
             let DefaultName(name) = expect_nonterminal(stream)?;
-            let value = match sudo_default(&name) {
-                Some(Setting::Flag(_)) => ConfigValue::Flag(false),
-                Some(Setting::List(_)) => ConfigValue::List(Mode::Set, vec![]),
-                Some(Setting::Text(OptTuple {
-                    negated: Some(val), ..
-                })) => ConfigValue::Text(val.map(|x| x.into())),
-                Some(Setting::Enum(OptTuple {
-                    negated: Some(val), ..
-                })) => ConfigValue::Enum(val),
-                Some(Setting::Integer(
-                    OptTuple {
-                        negated: Some(val), ..
-                    },
-                    _checker,
-                )) => ConfigValue::Num(val),
-                None => unrecoverable!(pos = value_pos, stream, "unknown setting: '{name}'"),
-                _ => unrecoverable!(
-                    pos = value_pos,
-                    stream,
-                    "'{name}' cannot be used in a boolean context"
-                ),
+            let Some(modifier) = defaults::negate(&name) else {
+                if defaults::set(&name).is_some() {
+                    unrecoverable!(pos = value_pos, stream, "unknown setting: '{name}'");
+                } else {
+                    unrecoverable!(
+                        pos = value_pos,
+                        stream,
+                        "'{name}' cannot be used in a boolean context"
+                    );
+                }
             };
-            make((name, value))
+
+            make(modifier)
         } else {
             let DefaultName(name) = try_nonterminal(stream)?;
-            let Some(cfg) = sudo_default(&name) else {
+            let Some(cfg) = defaults::set(&name) else {
                 unrecoverable!(pos = id_pos, stream, "unknown setting: '{name}'");
             };
 
             if is_syntax('+', stream)? {
-                list_items(Mode::Add, name, cfg, stream)
+                list_items(defaults::ListMode::Add, name, cfg, stream)
             } else if is_syntax('-', stream)? {
-                list_items(Mode::Del, name, cfg, stream)
+                list_items(defaults::ListMode::Del, name, cfg, stream)
             } else if is_syntax('=', stream)? {
                 let value_pos = stream.get_pos();
                 match cfg {
-                    Setting::Flag(_) => {
+                    defaults::SettingKind::Flag(_) => {
                         unrecoverable!(stream, "can't assign to boolean setting '{name}'")
                     }
-                    Setting::Integer(_, checker) => {
+                    defaults::SettingKind::Integer(checker) => {
                         let Numeric(denotation) = expect_nonterminal(stream)?;
-                        if let Some(value) = checker(&denotation) {
-                            make((name, ConfigValue::Num(value)))
+                        if let Some(modifier) = checker(&denotation) {
+                            make(modifier)
                         } else {
                             unrecoverable!(
                                 pos = value_pos,
@@ -719,34 +793,32 @@ impl Parse for (String, ConfigValue) {
                             );
                         }
                     }
-                    Setting::List(_) => {
+                    defaults::SettingKind::List(checker) => {
                         let items = parse_vars(stream)?;
-                        make((name, ConfigValue::List(Mode::Set, items)))
+
+                        make(checker(defaults::ListMode::Set, items))
                     }
-                    Setting::Text(_) => {
+                    defaults::SettingKind::Text(checker) => {
                         let text = text_item(stream)?;
-                        make((name, ConfigValue::Text(Some(text.into_boxed_str()))))
-                    }
-                    Setting::Enum(OptTuple { default: key, .. }) => {
-                        let text = text_item(stream)?;
-                        let Some(value) = key.alt(&text) else {
+                        let Some(modifier) = checker(&text) else {
                             unrecoverable!(
                                 pos = value_pos,
                                 stream,
                                 "'{text}' is not a valid value for {name}"
                             );
                         };
-                        make((name, ConfigValue::Enum(value)))
+                        make(modifier)
                     }
                 }
             } else {
-                if !matches!(cfg, Setting::Flag(_)) {
+                let defaults::SettingKind::Flag(modifier) = cfg else {
                     unrecoverable!(pos = id_pos, stream, "'{name}' is not a boolean setting");
-                }
-                make((name, ConfigValue::Flag(true)))
+                };
+
+                make(modifier)
             }
         }
     }
 }
 
-impl Many for (String, ConfigValue) {}
+impl Many for defaults::SettingsModifier {}
