@@ -11,7 +11,7 @@
 //!
 //! ```ignore
 //! impl<T: Parse> Parse for LinkedList<T> {
-//!     fn parse(stream: &mut impl CharStream) -> Parsed<LinkedList<T>> {
+//!     fn parse(stream: &mut CharStream) -> Parsed<LinkedList<T>> {
 //!         let x = try_nonterminal(stream)?;
 //!         let mut tail = if is_syntax('+', stream)? {
 //!             expect_nonterminal(stream)?
@@ -28,12 +28,17 @@
 /// Type holding a parsed object (or error information if parsing failed)
 pub type Parsed<T> = Result<T, Status>;
 
-pub type Position = std::ops::Range<(usize, usize)>;
+#[derive(Copy, Clone)]
+#[cfg_attr(test, derive(Debug, PartialEq))]
+pub struct Span {
+    pub start: (usize, usize),
+    pub end: (usize, usize),
+}
 
 #[cfg_attr(test, derive(Debug, PartialEq))]
 pub enum Status {
-    Fatal(Position, String), // not recoverable; stream in inconsistent state
-    Reject,                  // parsing failed by no input consumed
+    Fatal(Span, String), // not recoverable; stream in inconsistent state
+    Reject,              // parsing failed by no input consumed
 }
 
 pub fn make<T>(value: T) -> Parsed<T> {
@@ -46,14 +51,12 @@ pub fn reject<T>() -> Parsed<T> {
 
 macro_rules! unrecoverable {
     (pos=$pos:expr, $stream:ident, $($str:expr),*) => {
-        return Err(crate::sudoers::basic_parser::Status::Fatal($pos .. CharStream::get_pos($stream), format![$($str),*]))
-    };
-    ($stream:ident, $($str:expr),*) => {
-        {
-            let pos = CharStream::get_pos($stream);
-            return Err(crate::sudoers::basic_parser::Status::Fatal(pos .. pos, format![$($str),*]))
-        }
+        return Err(crate::sudoers::basic_parser::Status::Fatal(Span { start: $pos, end: CharStream::get_pos($stream)}, format![$($str),*]))
     };
+    ($stream:ident, $($str:expr),*) => {{
+        let pos = CharStream::get_pos($stream);
+        return Err(crate::sudoers::basic_parser::Status::Fatal(Span { start: pos, end: pos }, format![$($str),*]))
+    }};
     ($($str:expr),*) => {
         return Err(crate::basic_parser::Status::Fatal(Default::default(), format![$($str),*]))
     };
@@ -77,26 +80,11 @@ pub use super::char_stream::CharStream;
 /// advanced beyond the accepted part of the input. i.e. if some input is consumed the method
 /// *MUST* be producing a `Some` value.
 pub trait Parse {
-    fn parse(stream: &mut impl CharStream) -> Parsed<Self>
+    fn parse(stream: &mut CharStream) -> Parsed<Self>
     where
         Self: Sized;
 }
 
-/// Primitive function: accepts one character that satisfies `predicate`. This is used in the majority
-/// of all the other `Parse` implementations instead of interfacing with the iterator directly
-/// (this can facilitate an easy switch to a different method of stream representation in the future).
-/// Unlike `Parse` implementations this *does not* consume trailing whitespace.
-/// This function is modelled on `next_if` on `std::Peekable`.
-pub fn accept_if(predicate: impl Fn(char) -> bool, stream: &mut impl CharStream) -> Option<char> {
-    let c = stream.peek()?;
-    if predicate(c) {
-        stream.advance();
-        Some(c)
-    } else {
-        None
-    }
-}
-
 /// Structures representing whitespace (trailing whitespace can contain comments)
 #[cfg_attr(test, derive(PartialEq, Eq))]
 struct LeadingWhitespace;
@@ -110,8 +98,8 @@ struct Comment;
 /// Accept zero or more whitespace characters; fails if the whitespace is not "leading" to something
 /// (which can be used to detect end-of-input).
 impl Parse for LeadingWhitespace {
-    fn parse(stream: &mut impl CharStream) -> Parsed<Self> {
-        let eat_space = |stream: &mut _| accept_if(|c| "\t ".contains(c), stream);
+    fn parse(stream: &mut CharStream) -> Parsed<Self> {
+        let eat_space = |stream: &mut CharStream| stream.next_if(|c| "\t ".contains(c));
         while eat_space(stream).is_some() {}
 
         if stream.peek().is_some() {
@@ -126,14 +114,14 @@ impl Parse for LeadingWhitespace {
 /// always succeeds (unless some serious error occurs). This parser also accepts comments,
 /// since those can form part of trailing white space.
 impl Parse for TrailingWhitespace {
-    fn parse(stream: &mut impl CharStream) -> Parsed<Self> {
+    fn parse(stream: &mut CharStream) -> Parsed<Self> {
         loop {
             let _ = LeadingWhitespace::parse(stream); // don't propagate any errors
 
             // line continuations
-            if accept_if(|c| c == '\\', stream).is_some() {
+            if stream.eat_char('\\') {
                 // do the equivalent of expect_syntax('\n', stream)?, without recursion
-                if accept_if(|c| c == '\n', stream).is_none() {
+                if !stream.eat_char('\n') {
                     unrecoverable!(stream, "stray escape sequence")
                 }
             } else {
@@ -147,27 +135,31 @@ impl Parse for TrailingWhitespace {
 
 /// Parses a comment
 impl Parse for Comment {
-    fn parse(stream: &mut impl CharStream) -> Parsed<Self> {
-        accept_if(|c| c == '#', stream).ok_or(Status::Reject)?;
-        while accept_if(|c| c != '\n', stream).is_some() {}
+    fn parse(stream: &mut CharStream) -> Parsed<Self> {
+        if !stream.eat_char('#') {
+            return Err(Status::Reject);
+        }
+        stream.skip_to_newline();
         make(Comment {})
     }
 }
 
-fn skip_trailing_whitespace(stream: &mut impl CharStream) -> Parsed<()> {
+fn skip_trailing_whitespace(stream: &mut CharStream) -> Parsed<()> {
     TrailingWhitespace::parse(stream)?;
     make(())
 }
 
 /// Adheres to the contract of the [Parse] trait, accepts one character and consumes trailing whitespace.
-pub fn try_syntax(syntax: char, stream: &mut impl CharStream) -> Parsed<()> {
-    accept_if(|c| c == syntax, stream).ok_or(Status::Reject)?;
+pub fn try_syntax(syntax: char, stream: &mut CharStream) -> Parsed<()> {
+    if !stream.eat_char(syntax) {
+        return Err(Status::Reject);
+    }
     skip_trailing_whitespace(stream)?;
     make(())
 }
 
 /// Similar to [try_syntax], but aborts parsing if the expected character is not found.
-pub fn expect_syntax(syntax: char, stream: &mut impl CharStream) -> Parsed<()> {
+pub fn expect_syntax(syntax: char, stream: &mut CharStream) -> Parsed<()> {
     if try_syntax(syntax, stream).is_err() {
         let str = if let Some(c) = stream.peek() {
             c.to_string()
@@ -180,14 +172,14 @@ pub fn expect_syntax(syntax: char, stream: &mut impl CharStream) -> Parsed<()> {
 }
 
 /// Convenience function: usually try_syntax is called as a test criterion; if this returns true, the input was consumed.
-pub fn is_syntax(syntax: char, stream: &mut impl CharStream) -> Parsed<bool> {
+pub fn is_syntax(syntax: char, stream: &mut CharStream) -> Parsed<bool> {
     let result = maybe(try_syntax(syntax, stream))?;
     make(result.is_some())
 }
 
 /// Interface for working with types that implement the [Parse] trait; this allows parsing to use
 /// type inference. Use this instead of calling [Parse::parse] directly.
-pub fn try_nonterminal<T: Parse>(stream: &mut impl CharStream) -> Parsed<T> {
+pub fn try_nonterminal<T: Parse>(stream: &mut CharStream) -> Parsed<T> {
     let result = T::parse(stream)?;
     skip_trailing_whitespace(stream)?;
     make(result)
@@ -197,7 +189,7 @@ pub fn try_nonterminal<T: Parse>(stream: &mut impl CharStream) -> Parsed<T> {
 /// the given type or gives a fatal parse error if this did not succeed.
 use super::ast_names::UserFriendly;
 
-pub fn expect_nonterminal<T: Parse + UserFriendly>(stream: &mut impl CharStream) -> Parsed<T> {
+pub fn expect_nonterminal<T: Parse + UserFriendly>(stream: &mut CharStream) -> Parsed<T> {
     let begin_pos = stream.get_pos();
     match try_nonterminal(stream) {
         Err(Status::Reject) => {
@@ -228,21 +220,21 @@ pub trait Token: Sized {
 
 /// Implementation of the [Parse] trait for anything that implements [Token]
 impl<T: Token> Parse for T {
-    fn parse(stream: &mut impl CharStream) -> Parsed<Self> {
+    fn parse(stream: &mut CharStream) -> Parsed<Self> {
         fn accept_escaped<T: Token>(
             pred: fn(char) -> bool,
-            stream: &mut impl CharStream,
+            stream: &mut CharStream,
         ) -> Parsed<char> {
             const ESCAPE: char = '\\';
-            if T::ALLOW_ESCAPE && accept_if(|c| c == ESCAPE, stream).is_some() {
-                if let Some(c) = accept_if(T::escaped, stream) {
+            if T::ALLOW_ESCAPE && stream.eat_char(ESCAPE) {
+                if let Some(c) = stream.next_if(T::escaped) {
                     Ok(c)
                 } else if pred(ESCAPE) {
                     Ok(ESCAPE)
                 } else {
                     unrecoverable!(stream, "illegal escape sequence")
                 }
-            } else if let Some(c) = accept_if(pred, stream) {
+            } else if let Some(c) = stream.next_if(pred) {
                 Ok(c)
             } else {
                 reject()
@@ -267,21 +259,17 @@ impl<T: Token> Parse for T {
 
 /// Parser for `Option<T>` (this can be used to make the code more readable)
 impl<T: Parse> Parse for Option<T> {
-    fn parse(stream: &mut impl CharStream) -> Parsed<Self> {
+    fn parse(stream: &mut CharStream) -> Parsed<Self> {
         maybe(T::parse(stream))
     }
 }
 
 /// Parsing method for lists of items separated by a given character; this adheres to the contract of the [Parse] trait.
-#[allow(clippy::multiple_bound_locations)]
-pub(super) fn parse_list<T: Parse>(
+pub(super) fn parse_list<T: Parse + UserFriendly>(
     sep_by: char,
     max: usize,
-    stream: &mut impl CharStream,
-) -> Parsed<Vec<T>>
-where
-    T: Parse + UserFriendly,
-{
+    stream: &mut CharStream,
+) -> Parsed<Vec<T>> {
     let mut elems = Vec::new();
     elems.push(try_nonterminal(stream)?);
     while maybe(try_syntax(sep_by, stream))?.is_some() {
@@ -304,13 +292,13 @@ pub trait Many {
 /// Generic implementation for parsing multiple items of a type `T` that implements the [Parse] and
 /// [Many] traits.
 impl<T: Parse + Many + UserFriendly> Parse for Vec<T> {
-    fn parse(stream: &mut impl CharStream) -> Parsed<Self> {
+    fn parse(stream: &mut CharStream) -> Parsed<Self> {
         parse_list(T::SEP, T::LIMIT, stream)
     }
 }
 
 /// Entry point utility function; parse a `Vec<T>` but with fatal error recovery per line
-pub fn parse_lines<T, Stream: CharStream>(stream: &mut Stream) -> Vec<Parsed<T>>
+pub fn parse_lines<T>(stream: &mut CharStream) -> Vec<Parsed<T>>
 where
     T: Parse + UserFriendly,
 {
@@ -326,17 +314,17 @@ where
         result.push(item);
 
         let _ = maybe(Comment::parse(stream));
-        if accept_if(|c| c == '\n', stream).is_none() {
+        if !stream.eat_char('\n') {
             if parsed_item_ok {
                 let msg = if stream.peek().is_none() {
                     "missing line terminator at end of file"
                 } else {
                     "garbage at end of line"
                 };
-                let error = |stream: &mut Stream| unrecoverable!(stream, "{msg}");
+                let error = |stream: &mut CharStream| unrecoverable!(stream, "{msg}");
                 result.push(error(stream));
             }
-            while accept_if(|c| c != '\n', stream).is_some() {}
+            stream.skip_to_newline();
         }
     }
 
@@ -344,7 +332,7 @@ where
 }
 
 #[cfg(test)]
-fn expect_complete<T: Parse>(stream: &mut impl CharStream) -> Parsed<T> {
+fn expect_complete<T: Parse>(stream: &mut CharStream) -> Parsed<T> {
     let result = expect_nonterminal(stream)?;
     if let Some(c) = stream.peek() {
         unrecoverable!(stream, "garbage at end of line: {c}")
@@ -356,7 +344,7 @@ fn expect_complete<T: Parse>(stream: &mut impl CharStream) -> Parsed<T> {
 /// AST constructors by hand.
 #[cfg(test)]
 pub fn parse_string<T: Parse>(text: &str) -> Parsed<T> {
-    expect_complete(&mut text.chars().peekable())
+    expect_complete(&mut CharStream::new(text.chars()))
 }
 
 #[cfg(test)]
@@ -392,7 +380,7 @@ mod test {
 
     #[test]
     fn lines_test() {
-        let input = |text: &str| parse_lines(&mut text.chars().peekable());
+        let input = |text: &str| parse_lines(&mut CharStream::new(text.chars()));
 
         let s = |text: &str| Ok(text.to_string());
         assert_eq!(input("hello\nworld\n"), vec![s("hello"), s("world")]);
