@@ -10,32 +10,44 @@ mod entry;
 mod tokens;
 
 use std::collections::{HashMap, HashSet};
+use std::io;
 use std::path::{Path, PathBuf};
-use std::{io, mem};
 
 use crate::common::resolve::resolve_path;
+use crate::defaults;
 use crate::log::auth_warn;
-use crate::system::interface::{UnixGroup, UnixUser};
+use crate::system::interface::{GroupId, UnixGroup, UnixUser, UserId};
 use crate::system::{self, can_execute};
 use ast::*;
 use tokens::*;
 
+pub type Settings = defaults::Settings;
+pub use basic_parser::Span;
+
 /// How many nested include files do we allow?
 const INCLUDE_LIMIT: u8 = 128;
 
 /// Export some necessary symbols from modules
-pub use ast::TextEnum;
 pub struct Error {
     pub source: Option<PathBuf>,
-    pub location: Option<basic_parser::Position>,
+    pub location: Option<basic_parser::Span>,
     pub message: String,
 }
 
+/// A "Customiser" represents a "Defaults" setting that has 'late binding'; i.e.
+/// cannot be determined simply by reading a sudoers configuration. This is used
+/// for Defaults@host, Defaults:user, Defaults>runas and Defaults!cmd.
+///
+/// I.e. the Setting modifications in the second part of the tuple only apply for
+/// items explicitly matched by the first part of the tuple.
+type Customiser<Scope> = (Scope, Vec<defaults::SettingsModifier>);
+
 #[derive(Default)]
 pub struct Sudoers {
     rules: Vec<PermissionSpec>,
     aliases: AliasTable,
     settings: Settings,
+    customisers: CustomiserTable,
 }
 
 /// A structure that represents what the user wants to do
@@ -52,6 +64,7 @@ pub struct ListRequest<'a, User: UnixUser, Group: UnixGroup> {
 }
 
 #[derive(Default)]
+#[cfg_attr(test, derive(Clone))]
 pub struct Judgement {
     flags: Option<Tag>,
     settings: Settings,
@@ -59,10 +72,12 @@ pub struct Judgement {
 
 mod policy;
 
-pub use policy::{Authorization, AuthorizationAllowed, DirChange, Policy, PreJudgementPolicy};
+pub use policy::{AuthenticatingUser, Authentication, Authorization, DirChange, Restrictions};
 
 pub use self::entry::Entry;
 
+type MatchedCommand<'a> = (Option<&'a RunAs>, (Tag, &'a Spec<Command>));
+
 /// This function takes a file argument for a sudoers file and processes it.
 impl Sudoers {
     pub fn open(path: impl AsRef<Path>) -> Result<(Sudoers, Vec<Error>), io::Error> {
@@ -78,12 +93,65 @@ impl Sudoers {
         Ok(analyze(path.as_ref(), sudoers))
     }
 
+    fn specify_host_user_runas<User: UnixUser + PartialEq<User>>(
+        &mut self,
+        hostname: &system::Hostname,
+        requesting_user: &User,
+        target_user: &User,
+    ) {
+        let customisers = std::mem::take(&mut self.customisers.non_cmnd);
+
+        let host_matcher = &match_token(hostname);
+        let user_matcher = &match_user(requesting_user);
+        let runas_matcher = &match_user(target_user);
+
+        let host_aliases = get_aliases(&self.aliases.host, host_matcher);
+        let user_aliases = get_aliases(&self.aliases.user, user_matcher);
+        let runas_aliases = get_aliases(&self.aliases.runas, runas_matcher);
+
+        let match_scope = |scope| match scope {
+            ConfigScope::Generic => true,
+            ConfigScope::Host(list) => find_item(&list, host_matcher, &host_aliases).is_some(),
+            ConfigScope::User(list) => find_item(&list, user_matcher, &user_aliases).is_some(),
+            ConfigScope::RunAs(list) => find_item(&list, runas_matcher, &runas_aliases).is_some(),
+            ConfigScope::Command(_list) => {
+                unreachable!("command-specific defaults are filtered out")
+            }
+        };
+
+        for (scope, modifiers) in customisers {
+            if match_scope(scope) {
+                for modifier in modifiers {
+                    modifier(&mut self.settings);
+                }
+            }
+        }
+    }
+
+    fn specify_command(&mut self, command: &Path, arguments: &[String]) {
+        let customisers = std::mem::take(&mut self.customisers.cmnd);
+
+        let cmnd_matcher = &match_command((command, arguments));
+        let cmnd_aliases = get_aliases(&self.aliases.cmnd, cmnd_matcher);
+
+        for (scope, modifiers) in customisers {
+            if find_item(&scope, cmnd_matcher, &cmnd_aliases).is_some() {
+                for modifier in modifiers {
+                    modifier(&mut self.settings);
+                }
+            }
+        }
+    }
+
     pub fn check<User: UnixUser + PartialEq<User>, Group: UnixGroup>(
-        &self,
+        &mut self,
         am_user: &User,
         on_host: &system::Hostname,
         request: Request<User, Group>,
     ) -> Judgement {
+        self.specify_host_user_runas(on_host, am_user, request.user);
+        self.specify_command(request.command, request.arguments);
+
         // exception: if user is root or does not switch users, NOPASSWD is implied
         let skip_passwd =
             am_user.is_root() || (request.user == am_user && in_group(am_user, request.group));
@@ -97,7 +165,7 @@ impl Sudoers {
 
         Judgement {
             flags,
-            settings: self.settings.clone(), // this is wasteful, but in the future this will not be a simple clone and it avoids a lifetime
+            settings: self.settings.clone(),
         }
     }
 
@@ -143,12 +211,11 @@ impl Sudoers {
     ///
     /// the outer iterator are the `User_Spec`s; the inner iterator are the `Cmnd_Spec`s of
     /// said `User_Spec`s
-    fn matching_user_specs<'a: 'b + 'c, 'b: 'c, 'c, User: UnixUser + PartialEq<User>>(
+    fn matching_user_specs<'a, User: UnixUser + PartialEq<User>>(
         &'a self,
-        invoking_user: &'b User,
-        hostname: &'c system::Hostname,
-    ) -> impl Iterator<Item = impl Iterator<Item = (Option<&'a RunAs>, (Tag, &'a Spec<Command>))> + 'b>
-           + 'c {
+        invoking_user: &'a User,
+        hostname: &'a system::Hostname,
+    ) -> impl Iterator<Item = impl Iterator<Item = MatchedCommand<'a>>> {
         let Self { rules, aliases, .. } = self;
         let user_aliases = get_aliases(&aliases.user, &match_user(invoking_user));
         let host_aliases = get_aliases(&aliases.host, &match_token(hostname));
@@ -166,27 +233,24 @@ impl Sudoers {
             })
     }
 
-    /// returns `User_Spec`s that match `invoking_user` and `hostname` in a print-able format
     pub fn matching_entries<'a, User: UnixUser + PartialEq<User>>(
         &'a self,
-        invoking_user: &User,
-        hostname: &system::Hostname,
-    ) -> Vec<Entry<'a>> {
-        // NOTE this method MUST NOT perform any filtering that `Self::check` does not do to
-        // ensure `sudo $command` and `sudo --list` use the same permission checking logic
+        invoking_user: &'a User,
+        hostname: &'a system::Hostname,
+    ) -> impl Iterator<Item = Entry<'a>> {
         let user_specs = self.matching_user_specs(invoking_user, hostname);
 
-        let cmnd_aliases = unfold_alias_table(&self.aliases.cmnd);
-        let mut entries = vec![];
-        for cmd_specs in user_specs {
-            group_cmd_specs_per_runas(cmd_specs, &mut entries, &cmnd_aliases);
-        }
-
-        entries
+        user_specs.flat_map(|cmd_specs| group_cmd_specs_per_runas(cmd_specs, &self.aliases.cmnd))
     }
 
-    pub(crate) fn solve_editor_path(&self) -> Option<PathBuf> {
-        if self.settings.flags.contains("env_editor") {
+    pub(crate) fn solve_editor_path<User: UnixUser + PartialEq<User>>(
+        mut self,
+        on_host: &system::Hostname,
+        am_user: &User,
+        target_user: &User,
+    ) -> Option<PathBuf> {
+        self.specify_host_user_runas(on_host, am_user, target_user);
+        if self.settings.env_editor() {
             for key in ["SUDO_EDITOR", "VISUAL", "EDITOR"] {
                 if let Some(var) = std::env::var_os(key) {
                     let path = Path::new(&var);
@@ -208,53 +272,44 @@ impl Sudoers {
     }
 }
 
+// a `take_while` variant that does not consume the first non-matching item
+fn peeking_take_while<'a, T>(
+    iter: &'a mut std::iter::Peekable<impl Iterator<Item = T>>,
+    pred: impl Fn(&T) -> bool + 'a,
+) -> impl Iterator<Item = T> + 'a {
+    std::iter::from_fn(move || iter.next_if(&pred))
+}
+
 fn group_cmd_specs_per_runas<'a>(
     cmnd_specs: impl Iterator<Item = (Option<&'a RunAs>, (Tag, &'a Spec<Command>))>,
-    entries: &mut Vec<Entry<'a>>,
-    cmnd_aliases: &HashMap<&String, &'a Vec<Spec<Command>>>,
-) {
-    static EMPTY_RUNAS: RunAs = RunAs {
-        users: Vec::new(),
-        groups: Vec::new(),
-    };
+    cmnd_aliases: &'a VecOrd<Def<Command>>,
+) -> impl Iterator<Item = Entry<'a>> {
+    // `distribute_tags` will have given every spec a reference to the "runas specification"
+    // that applies to it. The output of sudo --list splits the CmndSpec list based on that:
+    // every line only has a single "runas" specifier. So we need to combine them for that.
+    //
+    // But sudo --list also outputs lines that are from different lines in the sudoers file on
+    // different lines in the output of sudo --list, so we cannot compare "by value". Luckily,
+    // once a RunAs is parsed, it will have a unique identifier in the form of its address.
+    let origin = |runas: Option<&RunAs>| runas.map(|r| r as *const _);
 
-    let mut runas = None;
-    let mut collected_specs = vec![];
+    let mut cmnd_specs = cmnd_specs.peekable();
 
-    for (new_runas, (tag, spec)) in cmnd_specs {
-        if let Some(new_runas) = new_runas {
-            if !collected_specs.is_empty() {
-                entries.push(Entry::new(
-                    runas.take().unwrap_or(&EMPTY_RUNAS),
-                    mem::take(&mut collected_specs),
-                ));
-            }
+    std::iter::from_fn(move || {
+        if let Some(&(cur_runas, _)) = cmnd_specs.peek() {
+            let specs = peeking_take_while(&mut cmnd_specs, |&(runas, _)| {
+                origin(runas) == origin(cur_runas)
+            });
 
-            runas = Some(new_runas);
-        }
-
-        let (negate, meta) = match spec {
-            Qualified::Allow(meta) => (false, meta),
-            Qualified::Forbid(meta) => (true, meta),
-        };
-
-        if let Meta::Alias(alias_name) = meta {
-            if let Some(specs) = cmnd_aliases.get(alias_name) {
-                // expand Cmnd_Alias
-                for spec in specs.iter() {
-                    let new_spec = if negate { spec.negate() } else { spec.as_ref() };
-
-                    collected_specs.push((tag.clone(), new_spec))
-                }
-            }
+            Some(Entry::new(
+                cur_runas,
+                specs.map(|x| x.1).collect(),
+                cmnd_aliases,
+            ))
         } else {
-            collected_specs.push((tag, spec.as_ref()));
+            None
         }
-    }
-
-    if !collected_specs.is_empty() {
-        entries.push(Entry::new(runas.unwrap_or(&EMPTY_RUNAS), collected_specs));
-    }
+    })
 }
 
 fn read_sudoers<R: io::Read>(mut reader: R) -> io::Result<Vec<basic_parser::Parsed<Sudo>>> {
@@ -264,7 +319,7 @@ fn read_sudoers<R: io::Read>(mut reader: R) -> io::Result<Vec<basic_parser::Pars
 
     use basic_parser::parse_lines;
     use char_stream::*;
-    Ok(parse_lines(&mut PeekableWithPos::new(buffer.chars())))
+    Ok(parse_lines(&mut CharStream::new(buffer.chars())))
 }
 
 fn open_sudoers(path: &Path) -> io::Result<Vec<basic_parser::Parsed<Sudo>>> {
@@ -277,27 +332,41 @@ fn open_subsudoers(path: &Path) -> io::Result<Vec<basic_parser::Parsed<Sudo>>> {
     read_sudoers(source)
 }
 
+// note: trying to DRY using GAT's is tempting but doesn't make the code any shorter
+
 #[derive(Default)]
-pub(super) struct AliasTable {
+struct AliasTable {
     user: VecOrd<Def<UserSpecifier>>,
-    host: VecOrd<Def<tokens::Hostname>>,
+    host: VecOrd<Def<Hostname>>,
     cmnd: VecOrd<Def<Command>>,
     runas: VecOrd<Def<UserSpecifier>>,
 }
 
+#[derive(Default)]
+struct CustomiserTable {
+    non_cmnd: Vec<Customiser<ConfigScope>>,
+    cmnd: Vec<Customiser<SpecList<Command>>>,
+}
+
 /// A vector with a list defining the order in which it needs to be processed
+struct VecOrd<T>(Vec<usize>, Vec<T>);
 
-type VecOrd<T> = (Vec<usize>, Vec<T>);
+impl<T> Default for VecOrd<T> {
+    fn default() -> Self {
+        VecOrd(Vec::default(), Vec::default())
+    }
+}
 
-fn elems<T>(vec: &VecOrd<T>) -> impl Iterator<Item = &T> {
-    vec.0.iter().map(|&i| &vec.1[i])
+impl<T> VecOrd<T> {
+    fn iter(&self) -> impl DoubleEndedIterator<Item = &T> + Clone {
+        self.0.iter().map(|&i| &self.1[i])
+    }
 }
 
 /// Check if the user `am_user` is allowed to run `cmdline` on machine `on_host` as the requested
 /// user/group. Not that in the sudoers file, later permissions override earlier restrictions.
 /// The `cmdline` argument should already be ready to essentially feed to an exec() call; or be
 /// a special command like 'sudoedit'.
-
 // This code is structure to allow easily reading the 'happy path'; i.e. as soon as something
 // doesn't match, we escape using the '?' mechanism.
 fn check_permission<User: UnixUser + PartialEq<User>, Group: UnixGroup>(
@@ -313,8 +382,6 @@ fn check_permission<User: UnixUser + PartialEq<User>, Group: UnixGroup>(
     let runas_user_aliases = get_aliases(&aliases.runas, &match_user(request.user));
     let runas_group_aliases = get_aliases(&aliases.runas, &match_group_alias(request.group));
 
-    // NOTE to ensure `sudo $command` and `sudo --list` behave the same, both this function and
-    // `Sudoers::matching_entries` must call this `matching_user_specs` method
     let matching_user_specs = sudoers.matching_user_specs(am_user, on_host).flatten();
 
     let allowed_commands = matching_user_specs.filter_map(|(runas, cmdspec)| {
@@ -339,19 +406,27 @@ fn check_permission<User: UnixUser + PartialEq<User>, Group: UnixGroup>(
 /// Process a raw parsed AST bit of RunAs + Command specifications:
 /// - RunAs specifications distribute over the commands that follow (until overridden)
 /// - Tags accumulate over the entire line
-
 fn distribute_tags(
     runas_cmds: &[(Option<RunAs>, CommandSpec)],
 ) -> impl Iterator<Item = (Option<&RunAs>, (Tag, &Spec<Command>))> {
     runas_cmds.iter().scan(
         (None, Default::default()),
-        |(mut last_runas, tag), (runas, CommandSpec(mods, cmd))| {
-            last_runas = runas.as_ref().or(last_runas);
+        |(last_runas, tag), (runas, CommandSpec(mods, cmd))| {
+            *last_runas = runas.as_ref().or(*last_runas);
             for f in mods {
                 f(tag);
             }
 
-            Some((last_runas, (tag.clone(), cmd)))
+            let this_tag = match cmd {
+                Qualified::Allow(Meta::All) if tag.env != EnvironmentControl::Nosetenv => Tag {
+                    // "ALL" has an implicit "SETENV" that doesn't distribute
+                    env: EnvironmentControl::Setenv,
+                    ..tag.clone()
+                },
+                _ => tag.clone(),
+            };
+
+            Some((*last_runas, (this_tag, cmd)))
         },
     )
 }
@@ -364,7 +439,6 @@ type FoundAliases = HashMap<String, bool>;
 /// Find an item matching a certain predicate in an collection (optionally attributed) list of
 /// identifiers; identifiers can be directly identifying, wildcards, and can either be positive or
 /// negative (i.e. preceeded by an even number of exclamation marks in the sudoers file)
-
 fn find_item<'a, Predicate, Iter, T: 'a>(
     items: Iter,
     matches: &Predicate,
@@ -377,11 +451,11 @@ where
 {
     let mut result = None;
     for item in items {
-        let (judgement, who) = match item.clone().to_inner() {
+        let (judgement, who) = match item.as_inner() {
             Qualified::Forbid(x) => (false, x),
             Qualified::Allow(x) => (true, x),
         };
-        let info = || item.to_info();
+        let info = || item.into_info();
         match who {
             Meta::All => result = judgement.then(info),
             Meta::Only(ident) if matches(ident) => result = judgement.then(info),
@@ -401,11 +475,11 @@ where
 }
 
 /// A interface to access optional "satellite data"
-trait WithInfo: Clone {
+trait WithInfo {
     type Item;
     type Info;
-    fn to_inner(self) -> Self::Item;
-    fn to_info(self) -> Self::Info;
+    fn as_inner(&self) -> Self::Item;
+    fn into_info(self) -> Self::Info;
 }
 
 /// A specific interface for `Spec<T>` --- we can't make a generic one;
@@ -413,20 +487,20 @@ trait WithInfo: Clone {
 impl<'a, T> WithInfo for &'a Spec<T> {
     type Item = &'a Spec<T>;
     type Info = ();
-    fn to_inner(self) -> &'a Spec<T> {
+    fn as_inner(&self) -> &'a Spec<T> {
         self
     }
-    fn to_info(self) {}
+    fn into_info(self) {}
 }
 
 /// A commandspec can be "tagged"
 impl<'a> WithInfo for (Tag, &'a Spec<Command>) {
     type Item = &'a Spec<Command>;
     type Info = Tag;
-    fn to_inner(self) -> &'a Spec<Command> {
+    fn as_inner(&self) -> &'a Spec<Command> {
         self.1
     }
-    fn to_info(self) -> Tag {
+    fn into_info(self) -> Tag {
         self.0
     }
 }
@@ -436,7 +510,7 @@ fn match_user(user: &impl UnixUser) -> impl Fn(&UserSpecifier) -> bool + '_ {
     move |spec| match spec {
         UserSpecifier::User(id) => match_identifier(user, id),
         UserSpecifier::Group(Identifier::Name(name)) => user.in_group_by_name(name.as_cstr()),
-        UserSpecifier::Group(Identifier::ID(num)) => user.in_group_by_gid(*num),
+        UserSpecifier::Group(Identifier::ID(num)) => user.in_group_by_gid(GroupId::new(*num)),
         _ => todo!(), // nonunix-groups, netgroups, etc.
     }
 }
@@ -447,8 +521,8 @@ fn in_group(user: &impl UnixUser, group: &impl UnixGroup) -> bool {
 
 fn match_group(group: &impl UnixGroup) -> impl Fn(&Identifier) -> bool + '_ {
     move |id| match id {
-        Identifier::ID(num) => group.as_gid() == *num,
-        Identifier::Name(name) => group.try_as_name().map_or(false, |s| name == s),
+        Identifier::ID(num) => group.as_gid() == GroupId::new(*num),
+        Identifier::Name(name) => group.try_as_name().is_some_and(|s| name == s),
     }
 }
 
@@ -481,13 +555,8 @@ fn match_command<'a>((cmd, args): (&'a Path, &'a [String])) -> (impl Fn(&Command
     }
 }
 
-fn unfold_alias_table<T>(table: &VecOrd<Def<T>>) -> HashMap<&String, &Vec<Qualified<Meta<T>>>> {
-    elems(table).map(|Def(id, list)| (id, list)).collect()
-}
-
 /// Find all the aliases that a object is a member of; this requires [sanitize_alias_table] to have run first;
 /// I.e. this function should not be "pub".
-
 fn get_aliases<Predicate, T>(table: &VecOrd<Def<T>>, pred: &Predicate) -> FoundAliases
 where
     Predicate: Fn(&T) -> bool,
@@ -496,7 +565,7 @@ where
     let all = Qualified::Allow(Meta::All);
 
     let mut set = HashMap::new();
-    for Def(id, list) in elems(table) {
+    for Def(id, list) in table.iter() {
         if find_item(list, &pred, &set).is_some() {
             set.insert(id.clone(), true);
         } else if find_item(once(&all).chain(list), &pred, &set).is_none() {
@@ -510,61 +579,10 @@ where
 }
 
 /// Code to map an ast::Identifier to the UnixUser trait
-
 fn match_identifier(user: &impl UnixUser, ident: &ast::Identifier) -> bool {
     match ident {
         Identifier::Name(name) => user.has_name(name),
-        Identifier::ID(num) => user.has_uid(*num),
-    }
-}
-
-#[derive(Clone)]
-pub struct Settings {
-    pub flags: HashSet<String>,
-    pub str_value: HashMap<String, Option<Box<str>>>,
-    pub enum_value: HashMap<String, TextEnum>,
-    pub int_value: HashMap<String, i64>,
-    pub list: HashMap<String, HashSet<String>>,
-}
-
-impl Default for Settings {
-    fn default() -> Self {
-        let mut this = Settings {
-            flags: Default::default(),
-            str_value: Default::default(),
-            enum_value: Default::default(),
-            int_value: Default::default(),
-            list: Default::default(),
-        };
-
-        use crate::defaults::{sudo_default, OptTuple, SudoDefault};
-        for key in crate::defaults::ALL_PARAMS.iter() {
-            match sudo_default(key).expect("internal error") {
-                SudoDefault::Flag(default) => {
-                    if default {
-                        this.flags.insert(key.to_string());
-                    }
-                }
-                SudoDefault::Text(OptTuple { default, .. }) => {
-                    this.str_value
-                        .insert(key.to_string(), default.map(|x| x.into()));
-                }
-                SudoDefault::Enum(OptTuple { default, .. }) => {
-                    this.enum_value.insert(key.to_string(), default);
-                }
-                SudoDefault::Integer(OptTuple { default, .. }, _) => {
-                    this.int_value.insert(key.to_string(), default);
-                }
-                SudoDefault::List(default) => {
-                    this.list.insert(
-                        key.to_string(),
-                        default.iter().map(|x| x.to_string()).collect(),
-                    );
-                }
-            }
-        }
-
-        this
+        Identifier::ID(num) => user.has_uid(UserId::new(*num)),
     }
 }
 
@@ -573,7 +591,6 @@ fn analyze(
     path: &Path,
     sudoers: impl IntoIterator<Item = basic_parser::Parsed<Sudo>>,
 ) -> (Sudoers, Vec<Error>) {
-    use ConfigValue::*;
     use Directive::*;
 
     let mut result: Sudoers = Default::default();
@@ -593,6 +610,7 @@ fn analyze(
     fn include(
         cfg: &mut Sudoers,
         parent: &Path,
+        span: Span,
         path: &Path,
         diagnostics: &mut Vec<Error>,
         count: &mut u8,
@@ -600,7 +618,7 @@ fn analyze(
         if *count >= INCLUDE_LIMIT {
             diagnostics.push(Error {
                 source: Some(parent.to_owned()),
-                location: None,
+                location: Some(span),
                 message: format!("include file limit reached opening '{}'", path.display()),
             })
         // FIXME: this will cause an error in `visudo` if we open a non-privileged sudoers file
@@ -621,7 +639,7 @@ fn analyze(
 
                     diagnostics.push(Error {
                         source: Some(parent.to_owned()),
-                        location: None,
+                        location: Some(span),
                         message,
                     })
                 }
@@ -643,31 +661,44 @@ fn analyze(
 
                     Sudo::Spec(permission) => cfg.rules.push(permission),
 
-                    Sudo::Decl(UserAlias(mut def)) => cfg.aliases.user.1.append(&mut def),
                     Sudo::Decl(HostAlias(mut def)) => cfg.aliases.host.1.append(&mut def),
-                    Sudo::Decl(CmndAlias(mut def)) => cfg.aliases.cmnd.1.append(&mut def),
+                    Sudo::Decl(UserAlias(mut def)) => cfg.aliases.user.1.append(&mut def),
                     Sudo::Decl(RunasAlias(mut def)) => cfg.aliases.runas.1.append(&mut def),
+                    Sudo::Decl(CmndAlias(mut def)) => cfg.aliases.cmnd.1.append(&mut def),
 
-                    Sudo::Decl(Defaults(params)) => {
-                        for (name, value) in params {
-                            set_default(cfg, name, value)
+                    Sudo::Decl(Defaults(params, scope)) => {
+                        if let ConfigScope::Command(specs) = scope {
+                            cfg.customisers.cmnd.push((
+                                specs
+                                    .into_iter()
+                                    .map(|spec| spec.map(|simple_command| (simple_command, None)))
+                                    .collect(),
+                                params,
+                            ));
+                        } else {
+                            cfg.customisers.non_cmnd.push((scope, params));
                         }
                     }
 
-                    Sudo::Include(path) => include(
+                    Sudo::Include(path, span) => include(
                         cfg,
                         cur_path,
+                        span,
                         &resolve_relative(cur_path, path),
                         diagnostics,
                         safety_count,
                     ),
 
-                    Sudo::IncludeDir(path) => {
+                    Sudo::IncludeDir(path, span) => {
                         if path.contains("%h") {
-                            diagnostics.push(Error{
-                                    source: Some(cur_path.to_owned()),
-                                    location: None,
-                                    message: format!("cannot open sudoers file {path}: percent escape %h in includedir is unsupported")});
+                            diagnostics.push(Error {
+                                source: Some(cur_path.to_owned()),
+                                location: Some(span),
+                                message: format!(
+                                    "cannot open sudoers file {path}: \
+                                     percent escape %h in includedir is unsupported"
+                                ),
+                            });
                             continue;
                         }
 
@@ -675,7 +706,7 @@ fn analyze(
                         let Ok(files) = std::fs::read_dir(&path) else {
                             diagnostics.push(Error {
                                 source: Some(cur_path.to_owned()),
-                                location: None,
+                                location: Some(span),
                                 message: format!("cannot open sudoers file {}", path.display()),
                             });
                             continue;
@@ -693,7 +724,14 @@ fn analyze(
                             .collect::<Vec<_>>();
                         safe_files.sort();
                         for file in safe_files {
-                            include(cfg, cur_path, file.as_ref(), diagnostics, safety_count)
+                            include(
+                                cfg,
+                                cur_path,
+                                span,
+                                file.as_ref(),
+                                diagnostics,
+                                safety_count,
+                            )
                         }
                     }
                 },
@@ -708,39 +746,6 @@ fn analyze(
         }
     }
 
-    fn set_default(cfg: &mut Sudoers, name: String, value: ConfigValue) {
-        match value {
-            Flag(value) => {
-                if value {
-                    cfg.settings.flags.insert(name);
-                } else {
-                    cfg.settings.flags.remove(&name);
-                }
-            }
-            List(mode, values) => {
-                let slot: &mut _ = cfg.settings.list.entry(name).or_default();
-                match mode {
-                    Mode::Set => *slot = values.into_iter().collect(),
-                    Mode::Add => slot.extend(values),
-                    Mode::Del => {
-                        for key in values {
-                            slot.remove(&key);
-                        }
-                    }
-                }
-            }
-            Text(value) => {
-                cfg.settings.str_value.insert(name, value);
-            }
-            Enum(value) => {
-                cfg.settings.enum_value.insert(name, value);
-            }
-            Num(value) => {
-                cfg.settings.int_value.insert(name, value);
-            }
-        }
-    }
-
     let mut diagnostics = vec![];
     process(&mut result, path, sudoers, &mut diagnostics, &mut 0);
 
@@ -756,7 +761,6 @@ fn analyze(
 /// Alias definition inin a Sudoers file can come in any order; and aliases can refer to other aliases, etc.
 /// It is much easier if they are presented in a "definitional order" (i.e. aliases that use other aliases occur later)
 /// At the same time, this is a good place to detect problems in the aliases, such as unknown aliases and cycles.
-
 fn sanitize_alias_table<T>(table: &Vec<Def<T>>, diagnostics: &mut Vec<Error>) -> Vec<usize> {
     fn remqualify<U>(item: &Qualified<U>) -> &U {
         match item {
@@ -787,7 +791,7 @@ fn sanitize_alias_table<T>(table: &Vec<Def<T>>, diagnostics: &mut Vec<Error>) ->
                 let Def(_, members) = &self.table[pos];
                 for elem in members {
                     let Meta::Alias(name) = remqualify(elem) else {
-                        break;
+                        continue;
                     };
                     let Some(dependency) = self.table.iter().position(|Def(id, _)| id == name)
                     else {
