@@ -1,6 +1,7 @@
 use std::ffi::CStr;
 
 use super::ast;
+use super::char_stream::CharStream;
 use super::*;
 use basic_parser::{parse_eval, parse_lines, parse_string};
 
@@ -20,16 +21,16 @@ impl UnixUser for Named {
         self.0 == name
     }
 
-    fn has_uid(&self, uid: u32) -> bool {
-        dummy_cksum(self.0) == uid
+    fn has_uid(&self, uid: UserId) -> bool {
+        UserId::new(dummy_cksum(self.0)) == uid
     }
 
     fn in_group_by_name(&self, name: &CStr) -> bool {
         self.has_name(name.to_str().unwrap())
     }
 
-    fn in_group_by_gid(&self, gid: u32) -> bool {
-        dummy_cksum(self.0) == gid
+    fn in_group_by_gid(&self, gid: GroupId) -> bool {
+        GroupId::new(dummy_cksum(self.0)) == gid
     }
 
     fn is_root(&self) -> bool {
@@ -38,8 +39,8 @@ impl UnixUser for Named {
 }
 
 impl UnixGroup for Named {
-    fn as_gid(&self) -> crate::system::interface::GroupId {
-        dummy_cksum(self.0)
+    fn as_gid(&self) -> GroupId {
+        GroupId::new(dummy_cksum(self.0))
     }
     fn try_as_name(&self) -> Option<&str> {
         Some(self.0)
@@ -57,7 +58,7 @@ macro_rules! request {
 
 macro_rules! sudoer {
     ($($e:expr),*) => {
-        parse_lines(&mut [$($e),*, ""].join("\n").chars().peekable())
+        parse_lines(&mut CharStream::new([$($e),*, ""].join("\n").chars()))
             .into_iter()
             .map(|x| Ok::<_,basic_parser::Status>(x.unwrap()))
     }
@@ -71,7 +72,7 @@ fn parse_line(s: &str) -> Sudo {
 
 /// Returns `None` if a syntax error is encountered
 fn try_parse_line(s: &str) -> Option<Sudo> {
-    parse_lines(&mut [s, ""].join("").chars().peekable())
+    parse_lines(&mut CharStream::new([s, ""].join("").chars()))
         .into_iter()
         .next()?
         .ok()
@@ -91,19 +92,19 @@ fn permission_test() {
 
     macro_rules! FAIL {
         ([$($sudo:expr),*], $user:expr => $req:expr, $server:expr; $command:expr) => {
-            let (Sudoers { rules,aliases,settings }, _) = analyze(Path::new("/etc/fakesudoers"), sudoer![$($sudo),*]);
+            let (Sudoers { rules,aliases,settings, customisers }, _) = analyze(Path::new("/etc/fakesudoers"), sudoer![$($sudo),*]);
             let cmdvec = $command.split_whitespace().map(String::from).collect::<Vec<_>>();
             let req = Request { user: $req.0, group: $req.1, command: &realpath(cmdvec[0].as_ref()), arguments: &cmdvec[1..].to_vec() };
-            assert_eq!(Sudoers { rules, aliases, settings }.check(&Named($user), &system::Hostname::fake($server), req).flags, None);
+            assert_eq!(Sudoers { rules, aliases, settings, customisers }.check(&Named($user), &system::Hostname::fake($server), req).flags, None);
         }
     }
 
     macro_rules! pass {
         ([$($sudo:expr),*], $user:expr => $req:expr, $server:expr; $command:expr $(=> [$($key:ident : $val:expr),*])?) => {
-            let (Sudoers { rules,aliases,settings }, _) = analyze(Path::new("/etc/fakesudoers"), sudoer![$($sudo),*]);
+            let (Sudoers { rules,aliases,settings, customisers }, _) = analyze(Path::new("/etc/fakesudoers"), sudoer![$($sudo),*]);
             let cmdvec = $command.split_whitespace().map(String::from).collect::<Vec<_>>();
             let req = Request { user: $req.0, group: $req.1, command: &realpath(cmdvec[0].as_ref()), arguments: &cmdvec[1..].to_vec() };
-            let result = Sudoers { rules, aliases, settings }.check(&Named($user), &system::Hostname::fake($server), req).flags;
+            let result = Sudoers { rules, aliases, settings, customisers }.check(&Named($user), &system::Hostname::fake($server), req).flags;
             assert!(!result.is_none());
             $(
                 let result = result.unwrap();
@@ -175,7 +176,7 @@ fn permission_test() {
     pass!(["user ALL=(ALL:ALL) /bin/foo"], "user" => request! { user, user }, "server"; "/bin/foo" => [authenticate: Authenticate::Nopasswd]);
     pass!(["user ALL=(ALL:ALL) /bin/foo"], "user" => request! { user, root }, "server"; "/bin/foo" => [authenticate: Authenticate::None]);
 
-    assert_eq!(Named("user").as_gid(), 1466);
+    assert_eq!(Named("user").as_gid(), GroupId::new(1466));
     pass!(["#1466 server=(ALL:ALL) ALL"], "user" => root(), "server"; "/bin/hello");
     pass!(["%#1466 server=(ALL:ALL) ALL"], "user" => root(), "server"; "/bin/hello");
     FAIL!(["#1466 server=(ALL:ALL) ALL"], "root" => root(), "server"; "/bin/hello");
@@ -214,6 +215,7 @@ fn permission_test() {
     pass!(["user ALL=(root) NOPASSWD: /bin/ls, (sudo) /bin/true"], "user" => request! { sudo }, "server"; "/bin/true" => [authenticate: Authenticate::Nopasswd]);
     FAIL!(["user ALL=(root) /bin/ls, (sudo) /bin/true"], "user" => request! { sudo }, "server"; "/bin/ls");
     FAIL!(["user ALL=(root) /bin/ls, (sudo) /bin/true"], "user" => request! { root }, "server"; "/bin/true");
+    pass!(["user ALL=(root) NOPASSWD: /bin/ls, (sudo) /bin/ls, /bin/true"], "user" => request! { sudo }, "server"; "/bin/true");
 
     SYNTAX!(["User_Alias, marc ALL = ALL"]);
 
@@ -247,26 +249,33 @@ fn permission_test() {
 
 #[test]
 fn default_bool_test() {
-    let (Sudoers { settings, .. }, _) = analyze(
+    let (mut sudoers, _) = analyze(
         Path::new("/etc/fakesudoers"),
         sudoer![
-            "Defaults env_reset",
+            "Defaults env_editor",
             "Defaults !use_pty",
+            "Defaults use_pty",
             "Defaults !env_keep",
             "Defaults !secure_path",
             "Defaults !env_editor"
         ],
     );
-    assert!(settings.flags.contains("env_reset"));
-    assert!(!settings.flags.contains("use_pty"));
-    assert!(settings.list["env_keep"].is_empty());
-    assert_eq!(settings.str_value["secure_path"], None);
-    assert!(!settings.flags.contains("env_editor"));
+    sudoers.specify_host_user_runas(
+        &system::Hostname::fake("host"),
+        &Named("user"),
+        &Named("root"),
+    );
+
+    assert!(!sudoers.settings.env_editor());
+    assert!(sudoers.settings.use_pty());
+    assert!(sudoers.settings.env_keep().is_empty());
+    assert_eq!(sudoers.settings.secure_path(), None);
+    assert!(!sudoers.settings.env_editor());
 }
 
 #[test]
 fn default_set_test() {
-    let (Sudoers { settings, .. }, _) = analyze(
+    let (mut sudoers, _) = analyze(
         Path::new("/etc/fakesudoers"),
         sudoer![
             "Defaults env_keep = \"FOO HUK BAR\"",
@@ -278,19 +287,25 @@ fn default_set_test() {
             "Defaults secure_path = /etc"
         ],
     );
+    sudoers.specify_host_user_runas(
+        &system::Hostname::fake("host"),
+        &Named("user"),
+        &Named("root"),
+    );
+
     assert_eq!(
-        settings.list["env_keep"],
-        ["FOO", "BAR"].into_iter().map(|x| x.to_string()).collect()
+        sudoers.settings.env_keep(),
+        &["FOO", "BAR"].into_iter().map(|x| x.to_string()).collect()
     );
     assert_eq!(
-        settings.list["env_check"],
-        ["FOO", "XYZZY"]
+        sudoers.settings.env_check(),
+        &["FOO", "XYZZY"]
             .into_iter()
             .map(|x| x.to_string())
             .collect()
     );
-    assert_eq!(settings.str_value["secure_path"].as_deref(), Some("/etc"));
-    assert_eq!(settings.int_value["passwd_tries"], 5);
+    assert_eq!(sudoers.settings.secure_path(), Some("/etc"));
+    assert_eq!(sudoers.settings.passwd_tries(), 5);
 
     assert!(parse_string::<Sudo>("Defaults verifypw = \"sometimes\"").is_err());
     assert!(parse_string::<Sudo>("Defaults verifypw = sometimes").is_err());
@@ -299,18 +314,24 @@ fn default_set_test() {
 
 #[test]
 fn default_multi_test() {
-    let (Sudoers { settings, .. }, _) = analyze(
+    let (mut sudoers, _) = analyze(
         Path::new("/etc/fakesudoers"),
         sudoer![
-        "Defaults env_reset, !use_pty, secure_path=/etc, env_keep = \"FOO BAR\", env_keep -= BAR"
+        "Defaults !env_editor, use_pty, secure_path=/etc, env_keep = \"FOO BAR\", env_keep -= BAR"
     ],
     );
-    assert!(settings.flags.contains("env_reset"));
-    assert!(!settings.flags.contains("use_pty"));
-    assert_eq!(settings.str_value["secure_path"].as_deref(), Some("/etc"));
+    sudoers.specify_host_user_runas(
+        &system::Hostname::fake("host"),
+        &Named("user"),
+        &Named("root"),
+    );
+
+    assert!(!sudoers.settings.env_editor());
+    assert!(sudoers.settings.use_pty());
+    assert_eq!(sudoers.settings.secure_path(), Some("/etc"));
     assert_eq!(
-        settings.list["env_keep"],
-        ["FOO".to_string()].into_iter().collect()
+        sudoers.settings.env_keep(),
+        &["FOO".to_string()].into_iter().collect()
     );
 }
 
@@ -384,7 +405,14 @@ fn gh676_percent_h_escape_unsupported() {
     assert_eq!(
         errs[0].message,
         "cannot open sudoers file /etc/%h: percent escape %h in includedir is unsupported"
-    )
+    );
+    assert_eq!(
+        errs[0].location,
+        Some(Span {
+            start: (1, 2),
+            end: (1, 23)
+        })
+    );
 }
 
 #[test]
@@ -412,6 +440,100 @@ fn defaults_regression() {
     assert!(try_parse_line("Defaults .mymachine=ALL").is_none())
 }
 
+#[test]
+fn specific_defaults() {
+    assert!(parse_line("Defaults !use_pty").is_decl());
+    assert!(try_parse_line("Defaults!use_pty").is_none());
+    assert!(parse_line("Defaults!/bin/bash !use_pty").is_decl());
+    assert!(try_parse_line("Defaults!/bin/bash!use_pty").is_none());
+    assert!(try_parse_line("Defaults !/bin/bash !use_pty").is_none());
+    assert!(try_parse_line("Defaults !/bin/bash").is_none());
+    assert!(parse_line("Defaults@host !use_pty").is_decl());
+    assert!(parse_line("Defaults@host!use_pty").is_decl());
+    assert!(try_parse_line("Defaults @host!use_pty").is_none());
+    assert!(try_parse_line("Defaults @host !use_pty").is_none());
+    assert!(parse_line("Defaults:user !use_pty").is_decl());
+    assert!(parse_line("Defaults:user!use_pty").is_decl());
+    assert!(try_parse_line("Defaults :user!use_pty").is_none());
+    assert!(try_parse_line("Defaults :user !use_pty").is_none());
+    assert!(parse_line("Defaults>user !use_pty").is_decl());
+    assert!(parse_line("Defaults>user!use_pty").is_decl());
+    assert!(try_parse_line("Defaults >user!use_pty").is_none());
+    assert!(try_parse_line("Defaults >user !use_pty").is_none());
+}
+
+#[test]
+fn default_specific_test() {
+    let sudoers = || {
+        analyze(
+            Path::new("/etc/fakesudoers"),
+            sudoer![
+                "Defaults!RR use_pty",
+                "Defaults env_editor",
+                "Defaults@host !env_editor",
+                "Defaults !use_pty",
+                "Defaults:user use_pty",
+                "Defaults !secure_path",
+                "Defaults>runas secure_path=\"/bin\"",
+                "Defaults!/bin/foo !env_keep",
+                "Cmnd_Alias RR=/usr/bin/rr twice"
+            ],
+        )
+    };
+
+    let (mut base_sudoers, _) = sudoers();
+    base_sudoers.specify_host_user_runas(
+        &system::Hostname::fake("generic"),
+        &Named("generic"),
+        &Named("generic"),
+    );
+
+    assert!(base_sudoers.settings.env_editor());
+    assert!(!base_sudoers.settings.use_pty());
+    assert!(base_sudoers.settings.env_keep().contains("COLORS"));
+    assert_eq!(base_sudoers.settings.secure_path(), None);
+
+    let (mut mod_sudoers, _) = sudoers();
+    mod_sudoers.specify_host_user_runas(
+        &system::Hostname::fake("host"),
+        &Named("user"),
+        &Named("root"),
+    );
+    assert!(!mod_sudoers.settings.env_editor());
+    assert!(mod_sudoers.settings.use_pty());
+    assert!(mod_sudoers.settings.env_keep().contains("COLORS"));
+    assert_eq!(mod_sudoers.settings.secure_path(), None);
+
+    let (mut mod_sudoers, _) = sudoers();
+    mod_sudoers.specify_host_user_runas(
+        &system::Hostname::fake("machine"),
+        &Named("admin"),
+        &Named("runas"),
+    );
+    assert!(mod_sudoers.settings.env_editor());
+    assert!(!mod_sudoers.settings.use_pty());
+    assert!(mod_sudoers.settings.env_keep().contains("COLORS"));
+    assert_eq!(mod_sudoers.settings.secure_path(), Some("/bin"));
+    mod_sudoers.specify_command(Path::new("/bin/foo"), &["".to_string(), "a".to_string()]);
+    assert!(mod_sudoers.settings.env_keep().is_empty());
+
+    let (mut mod_sudoers, _) = sudoers();
+    mod_sudoers.specify_host_user_runas(
+        &system::Hostname::fake("machine"),
+        &Named("admin"),
+        &Named("self"),
+    );
+    mod_sudoers.specify_command(Path::new("/usr/bin/rr"), &["thrice".to_string()]);
+    assert!(mod_sudoers.settings.env_editor());
+    assert!(!mod_sudoers.settings.use_pty());
+    assert!(mod_sudoers.settings.env_keep().contains("COLORS"));
+    assert_eq!(mod_sudoers.settings.secure_path(), None);
+
+    let (mut mod_sudoers, _) = sudoers();
+    mod_sudoers.specify_command(Path::new("/usr/bin/rr"), &["twice".to_string()]);
+    assert!(mod_sudoers.settings.use_pty());
+}
+
 #[test]
 fn useralias_underscore_regression() {
     let sudo = parse_line("FOO_BAR ALL=ALL");
@@ -422,6 +544,16 @@ fn useralias_underscore_regression() {
         .is_alias());
 }
 
+#[test]
+fn regression_check_recursion() {
+    let (_, error) = analyze(
+        Path::new("/etc/fakesudoers"),
+        sudoer!["User_Alias A=user, B", "User_Alias B=A"],
+    );
+
+    assert!(!error.is_empty());
+}
+
 fn test_topo_sort(n: usize) {
     let alias = |s: &str| Qualified::Allow(Meta::<UserSpecifier>::Alias(s.to_string()));
     let stop = || Qualified::Allow(Meta::<UserSpecifier>::All);
