@@ -1,5 +1,7 @@
 use core::fmt;
 // TODO: remove unused attribute when system is cleaned up
+#[cfg(target_os = "linux")]
+use std::str::FromStr;
 use std::{
     collections::BTreeSet,
     ffi::{c_uint, CStr, CString},
@@ -7,11 +9,10 @@ use std::{
     mem::MaybeUninit,
     ops,
     os::{
-        fd::AsRawFd,
+        fd::{AsFd, AsRawFd},
         unix::{self, prelude::OsStrExt},
     },
     path::{Path, PathBuf},
-    str::FromStr,
 };
 
 use crate::{
@@ -30,6 +31,8 @@ mod audit;
 // generalized traits for when we want to hide implementations
 pub mod interface;
 
+pub mod kernel;
+
 pub mod file;
 
 pub mod time;
@@ -47,10 +50,12 @@ pub(crate) fn can_execute<P: AsRef<Path>>(path: P) -> bool {
         return false;
     };
 
+    // SAFETY: we are passing a proper pointer to access
     unsafe { libc::access(path.as_ptr(), libc::X_OK) == 0 }
 }
 
 pub(crate) fn _exit(status: libc::c_int) -> ! {
+    // SAFETY: this function is safe to call
     unsafe { libc::_exit(status) }
 }
 
@@ -67,13 +72,20 @@ impl FileCloser {
         }
     }
 
-    pub(crate) fn except<F: AsRawFd>(&mut self, fd: &F) {
-        self.fds.insert(fd.as_raw_fd() as c_uint);
+    pub(crate) fn except<F: AsFd>(&mut self, fd: &F) {
+        self.fds.insert(fd.as_fd().as_raw_fd() as c_uint);
     }
 
     /// Close every file descriptor that is not one of the IO streams or one of the file
     /// descriptors passed via [`FileCloser::except`].
-    pub(crate) fn close_the_universe(self) -> io::Result<()> {
+    ///
+    /// # Safety
+    ///
+    /// Incorrect use of this method can violate [I/O Safety](https://doc.rust-lang.org/std/io/index.html#io-safety)
+    /// by closing fds that are not owned by `FileCloser`. The caller needs to ensure that none of
+    /// the closed fds are ever accessed again.
+    // FIXME do not return a Result. This makes it way to easy to accidentally violate the safety conditions.
+    pub(crate) unsafe fn close_the_universe(self) -> io::Result<()> {
         let mut fds = self.fds.into_iter();
 
         let Some(mut curr_fd) = fds.next() else {
@@ -104,7 +116,11 @@ impl FileCloser {
 
 fn close_range(min_fd: c_uint, max_fd: c_uint) -> io::Result<()> {
     if min_fd <= max_fd {
-        cerr(unsafe { libc::syscall(libc::SYS_close_range, min_fd, max_fd, 0 as c_uint) })?;
+        // SAFETY: this function is safe to call:
+        // - any errors while closing a specific fd will be effectively ignored
+        // - if the provided range or flags are invalid, that will be reported
+        //   as an error but will not cause undefined behaviour
+        cerr(unsafe { libc::close_range(min_fd, max_fd, 0) })?;
     }
 
     Ok(())
@@ -117,36 +133,62 @@ pub(crate) enum ForkResult {
     Child,
 }
 
-unsafe fn inner_fork() -> io::Result<ForkResult> {
+/// Create a new process.
+///
+/// # Safety
+///
+/// Must not be called in multithreaded programs.
+pub(crate) unsafe fn fork() -> io::Result<ForkResult> {
+    // FIXME add debug assertion that we are not currently using multiple threads.
+    // SAFETY: Calling async-signal-unsafe functions after fork is safe as the program is single
+    // threaded at this point according to the safety invariant of this function.
     let pid = cerr(unsafe { libc::fork() })?;
     if pid == 0 {
         Ok(ForkResult::Child)
     } else {
-        Ok(ForkResult::Parent(pid))
+        Ok(ForkResult::Parent(ProcessId::new(pid)))
     }
 }
 
-#[cfg(target_os = "linux")]
-/// Create a new process.
-pub(crate) fn fork() -> io::Result<ForkResult> {
-    // SAFETY: `fork` is implemented using `clone` in linux so we don't need to worry about signal
-    // safety.
-    unsafe { inner_fork() }
-}
-
-#[cfg(not(target_os = "linux"))]
-/// Create a new process.
+/// Create a new process with extra precautions for usage in tests.
 ///
 /// # Safety
 ///
 /// In a multithreaded program, only async-signal-safe functions are guaranteed to work in the
 /// child process until a call to `execve` or a similar function is done.
-pub(crate) unsafe fn fork() -> io::Result<ForkResult> {
-    inner_fork()
+#[cfg(test)]
+unsafe fn fork_for_test(child_func: impl FnOnce() -> std::convert::Infallible) -> ProcessId {
+    use std::io::Write;
+    use std::panic::{catch_unwind, AssertUnwindSafe};
+    use std::process::exit;
+
+    // SAFETY: Not really safe, but this is test only code.
+    match unsafe { fork() }.unwrap() {
+        ForkResult::Child => {
+            // Make sure that panics in the child always abort the process.
+            let err = match catch_unwind(AssertUnwindSafe(child_func)) {
+                Ok(res) => match res {},
+                Err(err) => err,
+            };
+
+            let s = if let Some(s) = err.downcast_ref::<&str>() {
+                s
+            } else if let Some(s) = err.downcast_ref::<String>() {
+                s
+            } else {
+                "Box<dyn Any>"
+            };
+            let _ = writeln!(std::io::stderr(), "{s}");
+
+            exit(101);
+        }
+        ForkResult::Parent(pid) => pid,
+    }
 }
 
 pub fn setsid() -> io::Result<ProcessId> {
-    cerr(unsafe { libc::setsid() })
+    // SAFETY: this function is memory-safe to call
+    Ok(ProcessId::new(cerr(unsafe { libc::setsid() })?))
 }
 
 #[derive(Clone)]
@@ -196,8 +238,10 @@ impl Hostname {
         let buffer_size = max_hostname_size + 1 /* null byte delimiter */ ;
         let mut buf = vec![0; buffer_size];
 
+        // SAFETY: we are passing a valid pointer to gethostname
         match cerr(unsafe { libc::gethostname(buf.as_mut_ptr(), buffer_size) }) {
             Ok(_) => Self {
+                // SAFETY: gethostname succeeded, so `buf` will hold a null-terminated C string
                 inner: unsafe { string_from_ptr(buf.as_ptr()) },
             },
 
@@ -216,6 +260,12 @@ pub fn syslog(priority: libc::c_int, facility: libc::c_int, message: &CStr) {
         Err(_) => panic!("syslog formatting string is not null-terminated"),
     };
 
+    // SAFETY:
+    // - "MSG" is a constant expression that is a null-terminated C string that represents "%s";
+    //   this also means that to achieve safety we MUST pass one more argument to syslog that is a proper
+    //   pointer to a null-terminated C string
+    // - message.as_ptr() is a pointer to a proper null-terminated C string (message being a &CStr)
+    // for more info: read the manpage for syslog(2)
     unsafe {
         libc::syslog(priority | facility, MSG, message.as_ptr());
     }
@@ -229,22 +279,35 @@ pub fn set_target_user(
 ) {
     use std::os::unix::process::CommandExt;
 
-    // add target group to list of additional groups if not present
-    if !target_user.groups.contains(&target_group.gid) {
-        target_user.groups.push(target_group.gid);
+    if let Some(index) = target_user
+        .groups
+        .iter()
+        .position(|id| id == &target_group.gid)
+    {
+        // make sure the requested group id is the first in the list (necessary on FreeBSD)
+        target_user.groups.swap(0, index)
+    } else {
+        // add target group to list of additional groups if not present
+        target_user.groups.insert(0, target_group.gid);
     }
 
     // we need to do this in a `pre_exec` call since the `groups` method in `process::Command` is unstable
     // see https://github.com/rust-lang/rust/blob/a01b4cc9f375f1b95fa8195daeea938d3d9c4c34/library/std/src/sys/unix/process/process_unix.rs#L329-L352
     // for the std implementation of the libc calls to `setgroups`, `setgid` and `setuid`
+    // SAFETY: Setuid, setgid and setgroups are async-signal-safe.
     unsafe {
         cmd.pre_exec(move || {
             cerr(libc::setgroups(
-                target_user.groups.len(),
-                target_user.groups.as_ptr(),
+                target_user.groups.len() as _,
+                // We can cast to gid_t because `GroupId` is marked as transparent
+                target_user.groups.as_ptr().cast::<libc::gid_t>(),
             ))?;
-            cerr(libc::setgid(target_group.gid))?;
-            cerr(libc::setuid(target_user.uid))?;
+            // setgid and setuid set the real, effective and saved version of the gid and uid
+            // respectively rather than just the real gid and uid. The original sudo uses setresgid
+            // and setresuid instead with all three arguments equal, but as this does the same as
+            // setgid and setuid using the latter is fine too.
+            cerr(libc::setgid(target_group.gid.inner()))?;
+            cerr(libc::setuid(target_user.uid.inner()))?;
 
             Ok(())
         });
@@ -255,30 +318,33 @@ pub fn set_target_user(
 pub fn kill(pid: ProcessId, signal: SignalNumber) -> io::Result<()> {
     // SAFETY: This function cannot cause UB even if `pid` is not a valid process ID or if
     // `signal` is not a valid signal code.
-    cerr(unsafe { libc::kill(pid, signal) }).map(|_| ())
+    cerr(unsafe { libc::kill(pid.inner(), signal) }).map(|_| ())
 }
 
 /// Send a signal to a process group with the specified ID.
 pub fn killpg(pgid: ProcessId, signal: SignalNumber) -> io::Result<()> {
     // SAFETY: This function cannot cause UB even if `pgid` is not a valid process ID or if
     // `signal` is not a valid signal code.
-    cerr(unsafe { libc::killpg(pgid, signal) }).map(|_| ())
+    cerr(unsafe { libc::killpg(pgid.inner(), signal) }).map(|_| ())
 }
 
 /// Get the process group ID of the current process.
 pub fn getpgrp() -> ProcessId {
-    unsafe { libc::getpgrp() }
+    // SAFETY: This function is always safe to call
+    ProcessId::new(unsafe { libc::getpgrp() })
 }
 
 /// Get a process group ID.
 pub fn getpgid(pid: ProcessId) -> io::Result<ProcessId> {
     // SAFETY: This function cannot cause UB even if `pid` is not a valid process ID
-    cerr(unsafe { libc::getpgid(pid) })
+    Ok(ProcessId::new(cerr(unsafe { libc::getpgid(pid.inner()) })?))
 }
 
 /// Set a process group ID.
 pub fn setpgid(pid: ProcessId, pgid: ProcessId) -> io::Result<()> {
-    cerr(unsafe { libc::setpgid(pid, pgid) }).map(|_| ())
+    // SAFETY: This function cannot cause UB even if `pid` or `pgid` are not a valid process IDs:
+    // https://pubs.opengroup.org/onlinepubs/007904975/functions/setpgid.html
+    cerr(unsafe { libc::setpgid(pid.inner(), pgid.inner()) }).map(|_| ())
 }
 
 pub fn chown<S: AsRef<CStr>>(
@@ -290,7 +356,9 @@ pub fn chown<S: AsRef<CStr>>(
     let uid = uid.into();
     let gid = gid.into();
 
-    cerr(unsafe { libc::chown(path, uid, gid) }).map(|_| ())
+    // SAFETY: path is a valid pointer to a null-terminated C string; chown cannot cause safety
+    // issues even if uid and/or gid would be invalid identifiers.
+    cerr(unsafe { libc::chown(path, uid.inner(), gid.inner()) }).map(|_| ())
 }
 
 #[derive(Debug, Clone, PartialEq)]
@@ -307,7 +375,7 @@ pub struct User {
 
 impl User {
     /// # Safety
-    /// This function expects `pwd` to be a result from a succesful call to `getpwXXX_r`.
+    /// This function expects `pwd` to be a result from a successful call to `getpwXXX_r`.
     /// (It can cause UB if any of `pwd`'s pointed-to strings does not have a null-terminator.)
     unsafe fn from_libc(pwd: &libc::passwd) -> Result<User, Error> {
         let mut buf_len: libc::c_int = 32;
@@ -315,6 +383,8 @@ impl User {
 
         while {
             groups_buffer = vec![0; buf_len as usize];
+            // SAFETY: getgrouplist is passed valid pointers
+            // in particular `groups_buffer` is an array of `buf.len()` bytes, as required
             let result = unsafe {
                 libc::getgrouplist(
                     pwd.pw_name,
@@ -337,16 +407,23 @@ impl User {
             panic!("invalid groups count returned from getgrouplist, this should not happen")
         });
 
-        Ok(User {
-            uid: pwd.pw_uid,
-            gid: pwd.pw_gid,
-            name: SudoString::new(string_from_ptr(pwd.pw_name))?,
-            gecos: string_from_ptr(pwd.pw_gecos),
-            home: SudoPath::new(os_string_from_ptr(pwd.pw_dir).into())?,
-            shell: os_string_from_ptr(pwd.pw_shell).into(),
-            passwd: string_from_ptr(pwd.pw_passwd),
-            groups: groups_buffer,
-        })
+        // SAFETY: All pointers were initialized by a successful call to `getpwXXX_r` as per the
+        // safety invariant of this function.
+        unsafe {
+            Ok(User {
+                uid: UserId::new(pwd.pw_uid),
+                gid: GroupId::new(pwd.pw_gid),
+                name: SudoString::new(string_from_ptr(pwd.pw_name))?,
+                gecos: string_from_ptr(pwd.pw_gecos),
+                home: SudoPath::new(os_string_from_ptr(pwd.pw_dir).into())?,
+                shell: os_string_from_ptr(pwd.pw_shell).into(),
+                passwd: string_from_ptr(pwd.pw_passwd),
+                groups: groups_buffer
+                    .iter()
+                    .map(|id| GroupId::new(*id))
+                    .collect::<Vec<_>>(),
+            })
+        }
     }
 
     pub fn from_uid(uid: UserId) -> Result<Option<User>, Error> {
@@ -354,9 +431,13 @@ impl User {
         let mut buf = vec![0; max_pw_size as usize];
         let mut pwd = MaybeUninit::uninit();
         let mut pwd_ptr = std::ptr::null_mut();
+        // SAFETY: getpwuid_r is passed valid (although partly uninitialized) pointers to memory,
+        // in particular `buf` points to an array of `buf.len()` bytes, as required.
+        // After this call, if `pwd_ptr` is not NULL, `*pwd_ptr` and `pwd` will be aliased;
+        // but we never dereference `pwd_ptr`.
         cerr(unsafe {
             libc::getpwuid_r(
-                uid,
+                uid.inner(),
                 pwd.as_mut_ptr(),
                 buf.as_mut_ptr(),
                 buf.len(),
@@ -366,37 +447,51 @@ impl User {
         if pwd_ptr.is_null() {
             Ok(None)
         } else {
+            // SAFETY: pwd_ptr was not null, and getpwuid_r succeeded, so we have assurances that
+            // the `pwd` structure was written to by getpwuid_r
             let pwd = unsafe { pwd.assume_init() };
+            // SAFETY: `pwd` was obtained by a call to getpwXXX_r, as required.
             unsafe { Self::from_libc(&pwd).map(Some) }
         }
     }
 
     pub fn effective_uid() -> UserId {
-        unsafe { libc::geteuid() }
+        // SAFETY: this function cannot cause memory safety issues
+        UserId::new(unsafe { libc::geteuid() })
     }
 
     pub fn effective_gid() -> GroupId {
-        unsafe { libc::getegid() }
+        // SAFETY: this function cannot cause memory safety issues
+        GroupId::new(unsafe { libc::getegid() })
     }
 
     pub fn real_uid() -> UserId {
-        unsafe { libc::getuid() }
+        // SAFETY: this function cannot cause memory safety issues
+        UserId::new(unsafe { libc::getuid() })
     }
 
     pub fn real_gid() -> GroupId {
-        unsafe { libc::getgid() }
+        // SAFETY: this function cannot cause memory safety issues
+        GroupId::new(unsafe { libc::getgid() })
     }
 
     pub fn real() -> Result<Option<User>, Error> {
         Self::from_uid(Self::real_uid())
     }
 
+    pub fn primary_group(&self) -> std::io::Result<Group> {
+        // Use from_gid_unchecked here to ensure that we can still resolve when
+        // the /etc/group entry for the primary group is missing.
+        Group::from_gid_unchecked(self.gid)
+    }
+
     pub fn from_name(name_c: &CStr) -> Result<Option<User>, Error> {
         let max_pw_size = sysconf(libc::_SC_GETPW_R_SIZE_MAX).unwrap_or(16_384);
         let mut buf = vec![0; max_pw_size as usize];
         let mut pwd = MaybeUninit::uninit();
         let mut pwd_ptr = std::ptr::null_mut();
 
+        // SAFETY: analogous to getpwuid_r above
         cerr(unsafe {
             libc::getpwnam_r(
                 name_c.as_ptr(),
@@ -409,7 +504,10 @@ impl User {
         if pwd_ptr.is_null() {
             Ok(None)
         } else {
+            // SAFETY: pwd_ptr was not null, and getpwnam_r succeeded, so we have assurances that
+            // the `pwd` structure was written to by getpwnam_r
             let pwd = unsafe { pwd.assume_init() };
+            // SAFETY: `pwd` was obtained by a call to getpwXXX_r, as required.
             unsafe { Self::from_libc(&pwd).map(Some) }
         }
     }
@@ -419,29 +517,34 @@ impl User {
 #[cfg_attr(test, derive(PartialEq))]
 pub struct Group {
     pub gid: GroupId,
-    pub name: String,
+    pub name: Option<String>,
 }
 
 impl Group {
     /// # Safety
-    /// This function expects `grp` to be a result from a succesful call to `getgrXXX_r`.
+    /// This function expects `grp` to be a result from a successful call to `getgrXXX_r`.
     /// In particular the grp.gr_mem pointer is assumed to be non-null, and pointing to a
     /// null-terminated list; the pointed-to strings are expected to be null-terminated.
     unsafe fn from_libc(grp: &libc::group) -> Group {
+        // SAFETY: The name pointer is initialized by a successful call to `getgrXXX_r` as per the
+        // safety invariant of this function.
+        let name = unsafe { string_from_ptr(grp.gr_name) };
         Group {
-            gid: grp.gr_gid,
-            name: string_from_ptr(grp.gr_name),
+            gid: GroupId::new(grp.gr_gid),
+            name: Some(name),
         }
     }
 
-    pub fn from_gid(gid: GroupId) -> std::io::Result<Option<Group>> {
+    /// Lookup group for gid without returning an error when a /etc/group entry is missing.
+    fn from_gid_unchecked(gid: GroupId) -> std::io::Result<Group> {
         let max_gr_size = sysconf(libc::_SC_GETGR_R_SIZE_MAX).unwrap_or(16_384);
         let mut buf = vec![0; max_gr_size as usize];
         let mut grp = MaybeUninit::uninit();
         let mut grp_ptr = std::ptr::null_mut();
+        // SAFETY: analogous to getpwuid_r above
         cerr(unsafe {
             libc::getgrgid_r(
-                gid,
+                gid.inner(),
                 grp.as_mut_ptr(),
                 buf.as_mut_ptr(),
                 buf.len(),
@@ -449,10 +552,23 @@ impl Group {
             )
         })?;
         if grp_ptr.is_null() {
+            Ok(Group { gid, name: None })
+        } else {
+            // SAFETY: grp_ptr was not null, and getgrgid_r succeeded, so we have assurances that
+            // the `grp` structure was written to by getgrgid_r
+            let grp = unsafe { grp.assume_init() };
+            // SAFETY: `pwd` was obtained by a call to getgrXXX_r, as required.
+            Ok(unsafe { Group::from_libc(&grp) })
+        }
+    }
+
+    pub fn from_gid(gid: GroupId) -> std::io::Result<Option<Group>> {
+        let group = Self::from_gid_unchecked(gid)?;
+        if group.name.is_none() {
+            // No entry in /etc/group
             Ok(None)
         } else {
-            let grp = unsafe { grp.assume_init() };
-            Ok(Some(unsafe { Group::from_libc(&grp) }))
+            Ok(Some(group))
         }
     }
 
@@ -461,6 +577,7 @@ impl Group {
         let mut buf = vec![0; max_gr_size as usize];
         let mut grp = MaybeUninit::uninit();
         let mut grp_ptr = std::ptr::null_mut();
+        // SAFETY: analogous to getpwuid_r above
         cerr(unsafe {
             libc::getgrnam_r(
                 name_c.as_ptr(),
@@ -473,7 +590,10 @@ impl Group {
         if grp_ptr.is_null() {
             Ok(None)
         } else {
+            // SAFETY: grp_ptr was not null, and getgrgid_r succeeded, so we have assurances that
+            // the `grp` structure was written to by getgrgid_r
             let grp = unsafe { grp.assume_init() };
+            // SAFETY: `pwd` was obtained by a call to getgrXXX_r, as required.
             Ok(Some(unsafe { Group::from_libc(&grp) }))
         }
     }
@@ -485,6 +605,7 @@ pub enum WithProcess {
 }
 
 impl WithProcess {
+    #[cfg(target_os = "linux")]
     fn to_proc_string(&self) -> String {
         match self {
             WithProcess::Current => "self".into(),
@@ -519,15 +640,15 @@ impl Process {
     pub fn process_id() -> ProcessId {
         // NOTE libstd casts the `i32` that `libc::getpid` returns into `u32`
         // here we cast it back into `i32` (`ProcessId`)
-        std::process::id() as ProcessId
+        ProcessId::new(std::process::id() as i32)
     }
 
     /// Return the parent process identifier for the current process
     pub fn parent_id() -> Option<ProcessId> {
         // NOTE libstd casts the `i32` that `libc::getppid` returns into `u32`
         // here we cast it back into `i32` (`ProcessId`)
-        let pid = unix::process::parent_id() as ProcessId;
-        if pid == 0 {
+        let pid = ProcessId::new(unix::process::parent_id() as i32);
+        if !pid.is_valid() {
             None
         } else {
             Some(pid)
@@ -536,14 +657,17 @@ impl Process {
 
     /// Get the session id for the current process
     pub fn session_id() -> ProcessId {
-        unsafe { libc::getsid(0) }
+        // SAFETY: this function is explicitly safe to call with argument 0,
+        // and more generally getsid will never cause memory safety issues.
+        ProcessId::new(unsafe { libc::getsid(0) })
     }
 
     /// Returns the device identifier of the TTY device that is currently
     /// attached to the given process
+    #[cfg(target_os = "linux")]
     pub fn tty_device_id(pid: WithProcess) -> std::io::Result<Option<DeviceId>> {
         // device id of tty is displayed as a signed integer of 32 bits
-        let data: i32 = read_proc_stat(pid, 6)?;
+        let data: i32 = read_proc_stat(pid, 6 /* tty_nr */)?;
         if data == 0 {
             Ok(None)
         } else {
@@ -552,13 +676,72 @@ impl Process {
             // int. We convert via u32 because a direct conversion to DeviceId
             // would use sign extension, which would result in a different bit
             // representation
-            Ok(Some(data as u32 as DeviceId))
+            Ok(Some(DeviceId::new(data as u64)))
+        }
+    }
+
+    /// Returns the device identifier of the TTY device that is currently
+    /// attached to the given process
+    #[cfg(target_os = "freebsd")]
+    pub fn tty_device_id(pid: WithProcess) -> std::io::Result<Option<DeviceId>> {
+        use std::ffi::c_void;
+        use std::ptr;
+
+        let mut ki_proc: Vec<libc::kinfo_proc> = Vec::with_capacity(1);
+
+        let pid = match pid {
+            WithProcess::Current => std::process::id() as i32,
+            WithProcess::Other(pid) => pid.inner(),
+        };
+
+        loop {
+            let mut size = ki_proc.capacity() * size_of::<libc::kinfo_proc>();
+            match cerr(unsafe {
+                libc::sysctl(
+                    [
+                        libc::CTL_KERN,
+                        libc::KERN_PROC,
+                        libc::KERN_PROC_PID,
+                        pid,
+                        size_of::<libc::kinfo_proc>() as i32,
+                        1,
+                    ]
+                    .as_ptr(),
+                    4,
+                    ki_proc.as_mut_ptr().cast::<c_void>(),
+                    &mut size,
+                    ptr::null(),
+                    0,
+                )
+            }) {
+                Ok(_) => {
+                    assert!(size >= size_of::<libc::kinfo_proc>());
+                    // SAFETY: The above sysctl has initialized at least `size` bytes. We have
+                    // asserted that this is at least a single element.
+                    unsafe {
+                        ki_proc.set_len(1);
+                    }
+                    break;
+                }
+                Err(e) if e.raw_os_error() == Some(libc::ENOMEM) => {
+                    // Vector not big enough. Grow it by 10% and try again.
+                    ki_proc.reserve(ki_proc.capacity() + (ki_proc.capacity() + 9) / 10);
+                }
+                Err(e) => return Err(e),
+            }
+        }
+
+        if ki_proc[0].ki_tdev == !0 {
+            Ok(None)
+        } else {
+            Ok(Some(DeviceId::new(ki_proc[0].ki_tdev)))
         }
     }
 
     /// Get the process starting time of a specific process
+    #[cfg(target_os = "linux")]
     pub fn starting_time(pid: WithProcess) -> io::Result<SystemTime> {
-        let process_start: u64 = read_proc_stat(pid, 21)?;
+        let process_start: u64 = read_proc_stat(pid, 21 /* start_time */)?;
 
         // the startime field is stored in ticks since the system start, so we need to know how many
         // ticks go into a second
@@ -575,9 +758,76 @@ impl Process {
             ((process_start % ticks_per_second) * (1_000_000_000 / ticks_per_second)) as i64,
         ))
     }
+
+    /// Get the process starting time of a specific process
+    #[cfg(target_os = "freebsd")]
+    pub fn starting_time(pid: WithProcess) -> io::Result<SystemTime> {
+        use std::ffi::c_void;
+        use std::ptr;
+
+        let mut ki_proc: Vec<libc::kinfo_proc> = Vec::with_capacity(1);
+
+        let pid = match pid {
+            WithProcess::Current => std::process::id() as i32,
+            WithProcess::Other(pid) => pid.inner(),
+        };
+
+        loop {
+            let mut size = ki_proc.capacity() * size_of::<libc::kinfo_proc>();
+            match cerr(unsafe {
+                libc::sysctl(
+                    [
+                        libc::CTL_KERN,
+                        libc::KERN_PROC,
+                        libc::KERN_PROC_PID,
+                        pid,
+                        size_of::<libc::kinfo_proc>() as i32,
+                        1,
+                    ]
+                    .as_ptr(),
+                    4,
+                    ki_proc.as_mut_ptr().cast::<c_void>(),
+                    &mut size,
+                    ptr::null(),
+                    0,
+                )
+            }) {
+                Ok(_) => {
+                    assert!(size >= size_of::<libc::kinfo_proc>());
+                    // SAFETY: The above sysctl has initialized at least `size` bytes. We have
+                    // asserted that this is at least a single element.
+                    unsafe {
+                        ki_proc.set_len(1);
+                    }
+                    break;
+                }
+                Err(e) if e.raw_os_error() == Some(libc::ENOMEM) => {
+                    // Vector not big enough. Grow it by 10% and try again.
+                    ki_proc.reserve(ki_proc.capacity() + (ki_proc.capacity() + 9) / 10);
+                }
+                Err(e) => return Err(e),
+            }
+        }
+
+        let ki_start = ki_proc[0].ki_start;
+        Ok(SystemTime::new(ki_start.tv_sec, ki_start.tv_usec * 1000))
+    }
 }
 
+/// Read the n-th field (with 0-based indexing) from `/proc/<pid>/self`.
+///
+/// See ["Table 1-4: Contents of the stat fields" of "The /proc
+/// Filesystem"][proc_stat_fields] in the Linux docs for all available fields.
+///
+/// IMPORTANT: the first two fields are not accessible with this routine.
+///
+/// [proc_stat_fields]: https://www.kernel.org/doc/html/latest/filesystems/proc.html#id10
+#[cfg(target_os = "linux")]
 fn read_proc_stat<T: FromStr>(pid: WithProcess, field_idx: isize) -> io::Result<T> {
+    // the first two fields are skipped by the code below, and we never need them,
+    // so no point in implementing code for it in this private function.
+    debug_assert!(field_idx >= 2);
+
     // read from a specific pid file, or use `self` to refer to our own process
     let pidref = pid.to_proc_string();
 
@@ -649,20 +899,31 @@ pub fn make_zeroed_sigaction() -> libc::sigaction {
     unsafe { std::mem::zeroed() }
 }
 
+#[cfg(all(test, target_os = "linux"))]
+pub(crate) const ROOT_GROUP_NAME: &str = "root";
+
+#[cfg(all(test, not(target_os = "linux")))]
+pub(crate) const ROOT_GROUP_NAME: &str = "wheel";
+
 #[cfg(test)]
 mod tests {
     use std::{
         io::{self, Read, Write},
-        os::{fd::AsRawFd, unix::net::UnixStream},
+        os::{
+            fd::{AsFd, AsRawFd},
+            unix::net::UnixStream,
+        },
         process::exit,
     };
 
     use libc::SIGKILL;
 
+    use crate::system::interface::{GroupId, ProcessId, UserId};
+
     use super::{
-        fork, getpgrp, setpgid,
+        fork_for_test, getpgrp, setpgid,
         wait::{Wait, WaitOptions},
-        ForkResult, Group, User, WithProcess,
+        Group, User, WithProcess, ROOT_GROUP_NAME,
     };
 
     pub(super) fn tempfile() -> std::io::Result<std::fs::File> {
@@ -683,16 +944,30 @@ mod tests {
 
     #[test]
     fn test_get_user_and_group_by_id() {
-        let fixed_users = &[(0, "root"), (1, "daemon")];
+        let fixed_users = &[
+            (UserId::ROOT, "root"),
+            (
+                User::from_name(cstr!("daemon")).unwrap().unwrap().uid,
+                "daemon",
+            ),
+        ];
         for &(id, name) in fixed_users {
             let root = User::from_uid(id).unwrap().unwrap();
-            assert_eq!(root.uid, id as libc::uid_t);
+            assert_eq!(root.uid, id);
             assert_eq!(root.name, name);
         }
-        for &(id, name) in fixed_users {
+
+        let fixed_groups = &[
+            (GroupId::new(0), ROOT_GROUP_NAME),
+            (
+                Group::from_name(cstr!("daemon")).unwrap().unwrap().gid,
+                "daemon",
+            ),
+        ];
+        for &(id, name) in fixed_groups {
             let root = Group::from_gid(id).unwrap().unwrap();
-            assert_eq!(root.gid, id as libc::gid_t);
-            assert_eq!(root.name, name);
+            assert_eq!(root.gid, id);
+            assert_eq!(root.name.unwrap(), name);
         }
     }
 
@@ -723,8 +998,8 @@ mod tests {
                     }
                 },
                 Group {
-                    name: name.to_string(),
-                    gid,
+                    name: Some(name.to_string()),
+                    gid: GroupId::new(gid),
                 }
             )
         }
@@ -752,23 +1027,29 @@ mod tests {
         use super::{getpgid, setpgid};
 
         let pgrp = getpgrp();
-        assert_eq!(getpgid(0).unwrap(), pgrp);
-        assert_eq!(getpgid(std::process::id() as i32).unwrap(), pgrp);
+        assert_eq!(getpgid(ProcessId::new(0)).unwrap(), pgrp);
+        assert_eq!(
+            getpgid(ProcessId::new(std::process::id() as i32)).unwrap(),
+            pgrp
+        );
 
-        match super::fork().unwrap() {
-            ForkResult::Child => {
+        let child_pid = unsafe {
+            super::fork_for_test(|| {
                 // wait for the parent.
-                std::thread::sleep(std::time::Duration::from_secs(1))
-            }
-            ForkResult::Parent(child_pid) => {
-                // The child should be in our process group.
-                assert_eq!(getpgid(child_pid).unwrap(), getpgid(0).unwrap(),);
-                // Move the child to its own process group
-                setpgid(child_pid, child_pid).unwrap();
-                // The process group of the child should have changed.
-                assert_eq!(getpgid(child_pid).unwrap(), child_pid);
-            }
-        }
+                std::thread::sleep(std::time::Duration::from_secs(1));
+                exit(0);
+            })
+        };
+
+        // The child should be in our process group.
+        assert_eq!(
+            getpgid(child_pid).unwrap(),
+            getpgid(ProcessId::new(0)).unwrap(),
+        );
+        // Move the child to its own process group
+        setpgid(child_pid, child_pid).unwrap();
+        // The process group of the child should have changed.
+        assert_eq!(getpgid(child_pid).unwrap(), child_pid);
     }
     #[test]
     fn kill_test() {
@@ -776,7 +1057,7 @@ mod tests {
             .arg("1")
             .spawn()
             .unwrap();
-        super::kill(child.id() as i32, SIGKILL).unwrap();
+        super::kill(ProcessId::new(child.id() as i32), SIGKILL).unwrap();
         assert!(!child.wait().unwrap().success());
     }
     #[test]
@@ -784,16 +1065,20 @@ mod tests {
         // Create a socket so the children write to it if they aren't terminated by `killpg`.
         let (mut rx, mut tx) = UnixStream::pair().unwrap();
 
-        let ForkResult::Parent(pid1) = fork().unwrap() else {
-            std::thread::sleep(std::time::Duration::from_secs(1));
-            tx.write_all(&[42]).unwrap();
-            exit(0);
+        let pid1 = unsafe {
+            fork_for_test(|| {
+                std::thread::sleep(std::time::Duration::from_secs(1));
+                tx.write_all(&[42]).unwrap();
+                exit(0);
+            })
         };
 
-        let ForkResult::Parent(pid2) = fork().unwrap() else {
-            std::thread::sleep(std::time::Duration::from_secs(1));
-            tx.write_all(&[42]).unwrap();
-            exit(0);
+        let pid2 = unsafe {
+            fork_for_test(|| {
+                std::thread::sleep(std::time::Duration::from_secs(1));
+                tx.write_all(&[42]).unwrap();
+                exit(0);
+            })
         };
 
         drop(tx);
@@ -811,36 +1096,39 @@ mod tests {
         );
     }
 
-    fn is_closed<F: AsRawFd>(fd: &F) -> bool {
-        crate::cutils::cerr(unsafe { libc::fcntl(fd.as_raw_fd(), libc::F_GETFD) })
+    fn is_closed<F: AsFd>(fd: &F) -> bool {
+        crate::cutils::cerr(unsafe { libc::fcntl(fd.as_fd().as_raw_fd(), libc::F_GETFD) })
             .is_err_and(|err| err.raw_os_error() == Some(libc::EBADF))
     }
 
     #[test]
     fn close_the_universe() {
-        let ForkResult::Parent(child_pid) = fork().unwrap() else {
-            let should_close =
-                std::fs::File::open(std::env::temp_dir().join("should_close.txt")).unwrap();
-            assert!(!is_closed(&should_close));
+        let child_pid = unsafe {
+            fork_for_test(|| {
+                let should_close =
+                    std::fs::File::create(std::env::temp_dir().join("should_close.txt")).unwrap();
+                assert!(!is_closed(&should_close));
 
-            let should_not_close =
-                std::fs::File::open(std::env::temp_dir().join("should_not_close.txt")).unwrap();
-            assert!(!is_closed(&should_not_close));
+                let should_not_close =
+                    std::fs::File::create(std::env::temp_dir().join("should_not_close.txt"))
+                        .unwrap();
+                assert!(!is_closed(&should_not_close));
 
-            let mut closer = super::FileCloser::new();
+                let mut closer = super::FileCloser::new();
 
-            closer.except(&should_not_close);
+                closer.except(&should_not_close);
 
-            closer.close_the_universe().unwrap();
+                closer.close_the_universe().unwrap();
 
-            assert!(is_closed(&should_close));
+                assert!(is_closed(&should_close));
 
-            assert!(!is_closed(&io::stdin()));
-            assert!(!is_closed(&io::stdout()));
-            assert!(!is_closed(&io::stderr()));
-            assert!(!is_closed(&should_not_close));
+                assert!(!is_closed(&io::stdin()));
+                assert!(!is_closed(&io::stdout()));
+                assert!(!is_closed(&io::stderr()));
+                assert!(!is_closed(&should_not_close));
 
-            exit(0)
+                exit(0)
+            })
         };
 
         let (_, status) = child_pid.wait(WaitOptions::new()).unwrap();
@@ -849,23 +1137,44 @@ mod tests {
 
     #[test]
     fn except_stdio_is_fine() {
-        let ForkResult::Parent(child_pid) = fork().unwrap() else {
-            let mut closer = super::FileCloser::new();
+        let child_pid = unsafe {
+            fork_for_test(|| {
+                let mut closer = super::FileCloser::new();
 
-            closer.except(&io::stdin());
-            closer.except(&io::stdout());
-            closer.except(&io::stderr());
+                closer.except(&io::stdin());
+                closer.except(&io::stdout());
+                closer.except(&io::stderr());
 
-            closer.close_the_universe().unwrap();
+                closer.close_the_universe().unwrap();
 
-            assert!(!is_closed(&io::stdin()));
-            assert!(!is_closed(&io::stdout()));
-            assert!(!is_closed(&io::stderr()));
+                assert!(!is_closed(&io::stdin()));
+                assert!(!is_closed(&io::stdout()));
+                assert!(!is_closed(&io::stderr()));
 
-            exit(0)
+                exit(0)
+            })
         };
 
         let (_, status) = child_pid.wait(WaitOptions::new()).unwrap();
         assert_eq!(status.exit_status(), Some(0));
     }
+
+    #[cfg(target_os = "linux")]
+    #[test]
+    fn proc_stat_test() {
+        use super::{read_proc_stat, Process, WithProcess::Current};
+        // The process can be '(uninterruptible) sleeping' or 'running': it looks like the state
+        // field of /proc/pid/stat will show the state for the main thread of the process rather
+        // than for the process as a whole.
+        let state = read_proc_stat::<char>(Current, 2).unwrap();
+        assert!("SDR".contains(state), "{state} is not S, D or R");
+        let parent = Process::parent_id().unwrap();
+        // field 3 is always the parent process
+        assert_eq!(
+            parent,
+            ProcessId::new(read_proc_stat::<i32>(Current, 3).unwrap())
+        );
+        // this next field should always be 0 (which precedes an important bit of info for us!)
+        assert_eq!(0, read_proc_stat::<i32>(Current, 20).unwrap());
+    }
 }
