@@ -1,23 +1,25 @@
+#![forbid(unsafe_code)]
+
 mod cli;
 mod help;
 
 use std::{
-    ffi::{CString, OsString},
     fs::{File, Permissions},
     io::{self, Read, Seek, Write},
-    os::unix::prelude::{MetadataExt, OsStringExt, PermissionsExt},
+    os::unix::prelude::{MetadataExt, PermissionsExt},
     path::{Path, PathBuf},
     process::Command,
 };
 
 use crate::{
-    sudo::diagnostic,
+    common::resolve::CurrentUser,
+    sudo::{candidate_sudoers_file, diagnostic},
     sudoers::Sudoers,
     system::{
         can_execute,
-        file::{Chown, FileLock},
+        file::{create_temporary_dir, Chown, FileLock},
         signal::{consts::*, register_handlers, SignalStream},
-        User,
+        Hostname, User,
     },
 };
 
@@ -64,7 +66,9 @@ pub fn main() {
 }
 
 fn check(file_arg: Option<&str>, perms: bool, owner: bool) -> io::Result<()> {
-    let sudoers_path = Path::new(file_arg.unwrap_or("/etc/sudoers"));
+    let sudoers_path = file_arg
+        .map(Path::new)
+        .unwrap_or_else(|| candidate_sudoers_file());
 
     let sudoers_file = File::open(sudoers_path)
         .map_err(|err| io_msg!(err, "unable to open {}", sudoers_path.display()))?;
@@ -121,7 +125,9 @@ fn check(file_arg: Option<&str>, perms: bool, owner: bool) -> io::Result<()> {
 }
 
 fn run(file_arg: Option<&str>, perms: bool, owner: bool) -> io::Result<()> {
-    let sudoers_path = Path::new(file_arg.unwrap_or("/etc/sudoers"));
+    let sudoers_path = file_arg
+        .map(Path::new)
+        .unwrap_or_else(|| candidate_sudoers_file());
 
     let (sudoers_file, existed) = if sudoers_path.exists() {
         let file = File::options().read(true).write(true).open(sudoers_path)?;
@@ -205,8 +211,22 @@ fn edit_sudoers_file(
     mut tmp_file: File,
     tmp_path: &Path,
 ) -> io::Result<()> {
+    let mut stderr = io::stderr();
+
     let mut editor_path = None;
     let mut sudoers_contents = Vec::new();
+
+    // Since visudo is meant to run as root, resolve shouldn't fail
+    let current_user: User = match CurrentUser::resolve() {
+        Ok(user) => user.into(),
+        Err(err) => {
+            writeln!(stderr, "visudo: cannot resolve : {err}")?;
+            return Ok(());
+        }
+    };
+
+    let host_name = Hostname::resolve();
+
     if existed {
         // If the sudoers file existed, read its contents and write them into the temporary file.
         sudoers_file.read_to_end(&mut sudoers_contents)?;
@@ -218,7 +238,7 @@ fn edit_sudoers_file(
         let (sudoers, errors) = Sudoers::read(sudoers_contents.as_slice(), sudoers_path)?;
 
         if errors.is_empty() {
-            editor_path = sudoers.solve_editor_path();
+            editor_path = sudoers.solve_editor_path(&host_name, &current_user, &current_user);
         }
     }
 
@@ -227,7 +247,6 @@ fn edit_sudoers_file(
         None => editor_path_fallback()?,
     };
 
-    let mut stderr = io::stderr();
     loop {
         Command::new(&editor_path)
             .arg("--")
@@ -252,8 +271,14 @@ fn edit_sudoers_file(
 
         writeln!(stderr, "The provided sudoers file format is not recognized or contains syntax errors. Please review:\n")?;
 
-        for crate::sudoers::Error { message, .. } in errors {
-            writeln!(stderr, "syntax error: {message}")?;
+        for crate::sudoers::Error {
+            message,
+            source,
+            location,
+        } in errors
+        {
+            let path = source.as_deref().unwrap_or(sudoers_path);
+            diagnostic::diagnostic!("syntax error: {message}", path @ location);
         }
 
         writeln!(stderr)?;
@@ -299,9 +324,18 @@ fn edit_sudoers_file(
 }
 
 fn editor_path_fallback() -> io::Result<PathBuf> {
-    let path = Path::new("/usr/bin/editor");
-    if can_execute(path) {
-        return Ok(path.to_owned());
+    let editors = if cfg!(target_os = "linux") {
+        &["/usr/bin/editor"][..]
+    } else if cfg!(target_os = "freebsd") {
+        &["/usr/bin/vi"]
+    } else {
+        unimplemented!("unknown default editor for target")
+    };
+    for &editor in editors {
+        let path = Path::new(editor);
+        if can_execute(path) {
+            return Ok(path.to_owned());
+        }
     }
 
     Err(io::Error::new(
@@ -309,17 +343,3 @@ fn editor_path_fallback() -> io::Result<PathBuf> {
         "cannot find text editor",
     ))
 }
-
-fn create_temporary_dir() -> io::Result<PathBuf> {
-    let template = cstr!("/tmp/sudoers-XXXXXX").to_owned();
-
-    let ptr = unsafe { libc::mkdtemp(template.into_raw()) };
-
-    if ptr.is_null() {
-        return Err(io::Error::last_os_error());
-    }
-
-    let path = OsString::from_vec(unsafe { CString::from_raw(ptr) }.into_bytes()).into();
-
-    Ok(path)
-}
