@@ -14,7 +14,7 @@ DATE=$(grep -m1 '^##' "$PROJECT_DIR"/CHANGELOG.md | grep -o '[0-9]\{4\}-[0-9]\{2
 # Build binaries
 docker build --pull --tag "$BUILDER_IMAGE_TAG" --file "$SCRIPT_DIR/Dockerfile-release" "$SCRIPT_DIR"
 docker run --rm --user "$(id -u):$(id -g)" -v "$PROJECT_DIR:/build" -w "/build" "$BUILDER_IMAGE_TAG" cargo clean
-docker run --rm --user "$(id -u):$(id -g)" -v "$PROJECT_DIR:/build" -w "/build" "$BUILDER_IMAGE_TAG" cargo build --release
+docker run --rm --user "$(id -u):$(id -g)" -v "$PROJECT_DIR:/build" -w "/build" "$BUILDER_IMAGE_TAG" cargo build --release --features pam-login
 
 # Generate man pages
 "$PROJECT_DIR/util/generate-docs.sh"
@@ -34,12 +34,15 @@ rm -rf "$target_sudo"
 set -x
 
 # Build sudo
+umask u=rwx,g=rx,o=rx
 mkdir -p "$target_dir_sudo/bin"
 mkdir -p "$target_dir_sudo/share/man/man8"
+mkdir -p "$target_dir_sudo/share/man/man5"
 cp "$PROJECT_DIR/target/release/sudo" "$target_dir_sudo/bin/sudo"
 cp "$PROJECT_DIR/target/release/visudo" "$target_dir_sudo/bin/visudo"
 cp "$PROJECT_DIR/target/docs/man/sudo.8" "$target_dir_sudo/share/man/man8/sudo.8"
 cp "$PROJECT_DIR/target/docs/man/visudo.8" "$target_dir_sudo/share/man/man8/visudo.8"
+cp "$PROJECT_DIR/target/docs/man/sudoers.5" "$target_dir_sudo/share/man/man5/sudoers.5"
 mkdir -p "$target_dir_sudo/share/doc/sudo-rs/sudo"
 cp "$PROJECT_DIR/README.md" "$target_dir_sudo/share/doc/sudo-rs/sudo/README.md"
 cp "$PROJECT_DIR/CHANGELOG.md" "$target_dir_sudo/share/doc/sudo-rs/sudo/CHANGELOG.md"
@@ -54,7 +57,7 @@ set -x
 chown -R root:root "$target_dir_sudo"
 chmod +xs "$target_dir_sudo/bin/sudo"
 chmod +x "$target_dir_sudo/bin/visudo"
-(cd $target_dir_sudo && tar --mtime="UTC $DATE 00:00:00" --use-compress-program='gzip -9n' -cpvf "$target_sudo" *)
+(cd $target_dir_sudo && tar --mtime="UTC $DATE 00:00:00" --sort=name --use-compress-program='gzip -9n' -cpvf "$target_sudo" *)
 EOF
 
 # Build su
@@ -75,7 +78,7 @@ set -eo pipefail
 set -x
 chown -R root:root "$target_dir_su"
 chmod +xs "$target_dir_su/bin/su"
-(cd $target_dir_su && tar --mtime="UTC $DATE 00:00:00" --use-compress-program='gzip -9n' -cpvf "$target_su" *)
+(cd $target_dir_su && tar --mtime="UTC $DATE 00:00:00" --sort=name --use-compress-program='gzip -9n' -cpvf "$target_su" *)
 EOF
 
 (cd $TARGET_DIR_BASE && sha256sum -b *-$SUDO_RS_VERSION.tar.gz > "$TARGET_DIR_BASE/SHA256SUMS")
